#
# Copyright (C) 2018-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Convert a HEALPix FITS file to multi-resolution UNIQ indexing from the more
common IMPLICIT indexing.
"""

from . import ArgumentParser, FileType


def parser():
    parser = ArgumentParser()
    parser.add_argument('input', metavar='INPUT.fits',
                        type=FileType('rb'), help='Input FITS file')
    parser.add_argument('output', metavar='OUTPUT.fits[.gz]',
                        type=FileType('wb'), help='Output FITS file')
    return parser


def main(args=None):
    with parser().parse_args(args) as args:
        import warnings
        from astropy.io import fits
        from ..io import read_sky_map, write_sky_map

        hdus = fits.open(args.input)
        ordering = hdus[1].header['ORDERING']
        expected_orderings = {'NESTED', 'RING'}
        if ordering not in expected_orderings:
            msg = 'Expected the FITS file {} to have ordering {}, but it is {}'
            warnings.warn(msg.format(
                args.input.name, ' or '.join(expected_orderings), ordering))
        table = read_sky_map(hdus, moc=True)
        write_sky_map(args.output.name, table)



#
# Copyright (C) 2018-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Convert a HEALPix FITS file from multi-resolution UNIQ indexing to the more
common IMPLICIT indexing.
"""

from . import ArgumentParser, FileType


def parser():
    parser = ArgumentParser()
    parser.add_argument('input', metavar='INPUT.fits[.gz]',
                        type=FileType('rb'), help='Input FITS file')
    parser.add_argument('output', metavar='OUTPUT.fits',
                        type=FileType('wb'), help='Output FITS file')
    parser.add_argument('--nside', type=int, help='Output HEALPix resolution')
    return parser


def main(args=None):
    with parser().parse_args(args) as args:
        import logging
        import warnings
        import astropy_healpix as ah
        from astropy.io import fits
        from ..io import read_sky_map, write_sky_map
        from ..bayestar import rasterize

        log = logging.getLogger()

        if args.nside is None:
            order = None
        else:
            order = ah.nside_to_level(args.nside)

        log.info('reading FITS file %s', args.input.name)
        hdus = fits.open(args.input)
        ordering = hdus[1].header['ORDERING']
        expected_ordering = 'NUNIQ'
        if ordering != expected_ordering:
            msg = 'Expected the FITS file {} to have ordering {}, but it is {}'
            warnings.warn(
                msg.format(args.input.name, expected_ordering, ordering))
        log.debug('converting original FITS file to Astropy table')
        table = read_sky_map(hdus, moc=True)
        log.debug('flattening HEALPix tree')
        table = rasterize(table, order=order)
        log.info('writing FITS file %s', args.output.name)
        write_sky_map(args.output.name, table, nest=True)
        log.debug('done')



# Copyright (C) 2017-2022  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import logging

from ligo.gracedb import rest
from ligo.lw import ligolw

from .base import EventSource
from .ligolw import LigoLWEventSource, _read_xml

__all__ = ('GraceDBEventSource',)

log = logging.getLogger('BAYESTAR')


def _has_psds(xmldoc):
    for elem in xmldoc.getElementsByTagName(ligolw.LIGO_LW.tagName):
        if elem.hasAttribute('Name') and elem.Name == 'REAL8FrequencySeries':
            return True
    return False


class GraceDBEventSource(EventSource):
    """Read events from GraceDB.

    Parameters
    ----------
    graceids : list
        List of GraceDB ID strings.
    client : `ligo.gracedb.rest.GraceDb`, optional
        Client object

    Returns
    -------
    `~ligo.skymap.io.events.EventSource`

    """

    def __init__(self, graceids, client=None):
        if client is None:
            client = rest.GraceDb()
        self._client = client
        self._graceids = graceids

    def __iter__(self):
        return iter(self._graceids)

    def __getitem__(self, graceid):
        coinc_file, _ = _read_xml(self._client.files(graceid, 'coinc.xml'))
        if _has_psds(coinc_file):
            psd_file = coinc_file
        else:
            log.warning('The coinc.xml should contain a PSD, but it does not. '
                        'Attempting to download psd.xml.gz.')
            psd_file = self._client.files(graceid, 'psd.xml.gz')
        event, = LigoLWEventSource(
            coinc_file, psd_file=psd_file, coinc_def=None).values()
        return event

    def __len__(self):
        return len(self._graceids)


open = GraceDBEventSource



from contextlib import ExitStack
import errno
import gzip
import os
from pathlib import Path
import re
import subprocess

from ligo.lw.utils import load_filename
import h5py
import numpy as np
import pytest

from .. import events
from ..events.ligolw import ContentHandler
from ...util import sqlite

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')


@pytest.fixture
def mock_gracedb(monkeypatch):
    stack = ExitStack()

    class MockGraceDb(ExitStack):

        def files(self, graceid, filename):
            path = os.path.join(DATA_PATH, '{}_{}'.format(graceid, filename))
            try:
                return stack.enter_context(open(path, 'rb'))
            except IOError as e:
                if e.errno != errno.ENOENT:
                    raise
                return stack.enter_context(gzip.GzipFile(path + '.gz', 'rb'))

    with stack:
        monkeypatch.setattr('ligo.gracedb.rest.GraceDb', MockGraceDb)
        yield


def raises(expected_exception, msg):
    return pytest.raises(expected_exception, match='^' + re.escape(msg) + '$')


def ligolw_assertions(source):
    """Test common assertions for test_ligolw and test_sqlite."""
    assert len(source) == 250
    event = source[821759]
    assert len(event.singles) == 2
    assert event.singles[0].snr_series is event.singles[1].snr_series is None
    assert event.singles[0].detector == 'H1'
    assert event.singles[0].snr == 8.5362396
    assert event.singles[0].phase == -0.81192881
    assert (event.singles[0].time == event.singles[0].zerolag_time ==
            967328914.866107842)
    psd = event.singles[0].psd
    assert psd.f0 == 0.0
    assert psd.deltaF == 0.125
    assert event.singles[1].detector == 'L1'
    assert event.singles[1].snr == 10.36818
    assert event.singles[1].phase == 1.9740163
    assert (event.singles[1].time == event.singles[1].zerolag_time ==
            967328914.866513726)
    psd = event.singles[0].psd
    assert psd.f0 == 0.0
    assert psd.deltaF == 0.125
    assert event.template_args == {
        'mass1': 1.56687,
        'mass2': 1.474779,
        'spin1x': 0.0,
        'spin1y': 0.0,
        'spin1z': 0.0,
        'spin2x': 0.0,
        'spin2y': 0.0,
        'spin2z': 0.0,
        'f_final': 2047.0}


def test_unknown():
    with pytest.raises(IOError, match='Unknown file format'):
        events.open(__file__)


def test_ligolw():
    """Test reading events from LIGO-LW XML files."""
    source = events.open(os.path.join(DATA_PATH, '2016_subset.xml.gz'))
    ligolw_assertions(source)

    assert str(source).startswith('<LigoLWEventSource({288172: <LigoLWEvent(')
    assert repr(source).startswith('LigoLWEventSource({288172: <LigoLWEvent(')
    event, *_ = source.values()
    assert str(event).startswith(
        "<LigoLWEvent(singles=(<LigoLWSingleEvent(detector='H1', snr=12.")
    assert repr(event).startswith(
        "<LigoLWEvent(singles=(<LigoLWSingleEvent(detector='H1', snr=12.")
    single_event, *_ = event.singles
    assert str(single_event) == "<LigoLWSingleEvent(detector='H1', snr=12.035994, phase=-1.0371021, time=970976257.4808338)>"  # noqa
    assert repr(single_event) == "<LigoLWSingleEvent(detector='H1', snr=12.035994, phase=-1.0371021, time=970976257.4808338)>"  # noqa


def test_ligolw_document():
    """Test reading events from LIGO-LW XML document."""
    xmldoc = load_filename(
        os.path.join(DATA_PATH, '2016_subset.xml.gz'),
        contenthandler=ContentHandler)
    source = events.open(xmldoc, fallbackpath=DATA_PATH)
    ligolw_assertions(source)


def test_ligolw_pathlib():
    """Test reading file provided with Pathlib instead of string"""
    source = events.open(Path(DATA_PATH, '2016_subset.xml.gz'))
    ligolw_assertions(source)


def test_sqlite(tmpdir):
    """Test reading events from SQLite files."""
    # Convert the test data to SQLite format in the temporary directory.
    xmlfilename = os.path.join(DATA_PATH, '2016_subset.xml.gz')
    dbfilename = str(tmpdir / '2016_subset.sqlite')
    subprocess.check_call(['ligolw_sqlite', '-p',
                           xmlfilename, '-d', dbfilename])

    # Symbolicly link the PSD directory into the temporary directory.
    os.symlink(os.path.join(DATA_PATH, 'gstlal_reference_psd'),
               str(tmpdir / 'gstlal_reference_psd'))

    source = events.open(dbfilename)
    ligolw_assertions(source)

    with open(dbfilename, 'rb') as f:
        source = events.open(f)
        ligolw_assertions(source)

    with sqlite.open(dbfilename, 'r') as db:
        source = events.open(db)
        ligolw_assertions(source)


def test_gracedb(mock_gracedb):
    """Test reading events from GraceDB records."""
    source = events.gracedb.open(['G211117', 'G197392'])
    assert len(source) == 2
    for i, (event_id, event) in enumerate(source.items()):
        if i == 0:
            assert event_id == 'G211117'
            assert (event.singles[0].snr_series is event.singles[1].snr_series
                    is None)
            assert event.singles[0].detector == 'H1'
            assert event.singles[0].snr == 9.0802174
            assert event.singles[0].phase == -0.13969257
            assert (event.singles[0].time == event.singles[0].zerolag_time ==
                    1135136350.647757924)
            psd = event.singles[0].psd
            assert psd.f0 == 0.0
            assert psd.deltaF == 0.125
            assert event.singles[1].detector == 'L1'
            assert event.singles[1].snr == 7.3947201
            assert event.singles[1].phase == -2.7356486
            assert (event.singles[1].time == event.singles[1].zerolag_time ==
                    1135136350.646883043)
            psd = event.singles[1].psd
            assert psd.f0 == 0.0
            assert psd.deltaF == 0.125
            assert event.template_args == {
                'mass1': 19.924686,
                'mass2': 6.4254546,
                'spin1x': 0.0,
                'spin1y': 0.0,
                'spin1z': 0.33962944,
                'spin2x': 0.0,
                'spin2y': 0.0,
                'spin2z': -0.1238557,
                'f_final': 1024.0}
        elif i == 1:
            assert event_id == 'G197392'
            assert (event.singles[0].snr_series is event.singles[1].snr_series
                    is None)
            assert event.singles[0].detector == 'H1'
            assert event.singles[0].snr == 6.9068823
            assert event.singles[0].phase == 1.8298783
            assert (event.singles[0].time == event.singles[0].zerolag_time ==
                    1128678900.444335938)
            psd = event.singles[0].psd
            assert psd.f0 == 30.0
            assert psd.deltaF == 0.125
            assert event.singles[1].detector == 'L1'
            assert event.singles[1].snr == 6.8389997
            assert event.singles[1].phase == -1.0297496
            assert (event.singles[1].time == event.singles[1].zerolag_time ==
                    1128678900.445068359)
            psd = event.singles[1].psd
            assert psd.f0 == 30.0
            assert psd.deltaF == 0.125
            assert event.template_args == {
                'mass1': 32.064007,
                'mass2': 14.607587,
                'spin1x': 0.0,
                'spin1y': 0.0,
                'spin1z': 0.34881824,
                'spin2x': 0.0,
                'spin2y': 0.0,
                'spin2z': -0.53029484,
                'f_final': 0.0}


def test_detector_disabled(mock_gracedb):
    """Test reading from event sources with certain detectors disabled."""
    graceids = ('G211117', 'G197392')
    base_source = events.gracedb.open(graceids)

    source = events.detector_disabled.open(base_source, ['H1'])
    assert len(source) == 2
    for graceid, (event_id, event) in zip(graceids, source.items()):
        assert event_id == graceid
        assert len(event.singles) == 1
        assert event.singles[0].detector == 'L1'

    for event, base_event in zip(source.values(), base_source.values()):
        assert event.template_args == base_event.template_args

    # Now test that exceptions are raised when they are called for.
    expected_message = ('Disabling detectors {H1, L1} would have no effect on '
                        'this event with detectors {H1 L1}')
    nonraising_source = events.detector_disabled.open(
        base_source, ['H1, L1'], raises=False)
    raising_source = events.detector_disabled.open(
        base_source, ['H1, L1'])
    for event in nonraising_source.values():
        event.singles
    for event in raising_source.values():
        with raises(events.DetectorDisabledError, expected_message):
            event.singles

    # Now test that exceptions are raised when they are called for.
    expected_message = ('Disabling detectors {H1 L1 V1} would exclude all '
                        'data for this event with detectors {H1 L1}')
    nonraising_source = events.detector_disabled.open(
        base_source, ['H1', 'L1', 'V1'], raises=False)
    raising_source = events.detector_disabled.open(
        base_source, ['H1', 'L1', 'V1'])
    for event in nonraising_source.values():
        event.singles
    for event in raising_source.values():
        with raises(events.DetectorDisabledError, expected_message):
            event.singles


def test_hdf(tmpdir):
    """Test reading events from HDF5 files."""
    # Create test input files
    ifos = ['L1', 'H1']
    filenames = []
    filename = str(tmpdir / 'coincs.hdf')
    filenames.append(filename)
    with h5py.File(filename, 'w') as coinc_file:
        coinc_file.attrs['timeslide_interval'] = np.e
        coinc_group = coinc_file.create_group('foreground')
        coinc_group['template_id'] = np.arange(5)
        coinc_group['timeslide_id'] = np.arange(0, 50, 10)

        filename = str(tmpdir / 'H1L1-BANK.hdf')
        filenames.append(filename)
        with h5py.File(filename, 'w') as bank_file:
            bank_file.attrs['parameters'] = []

        for i, ifo in enumerate(ifos):
            coinc_file.attrs['detector_{}'.format(i + 1)] = ifo
            coinc_group['trigger_id{}'.format(i + 1)] = np.arange(5)

            filename = str(tmpdir / (ifo + '_triggers.hdf'))
            filenames.append(filename)
            with h5py.File(filename, 'w') as trigger_file:
                trigger_group = trigger_file.create_group(ifo)
                trigger_group['snr'] = i + np.arange(5) * np.pi
                trigger_group['coa_phase'] = i + np.arange(5) * np.pi**2
                trigger_group['end_time'] = i + np.arange(5) * np.pi**3

            filename = str(tmpdir / (ifo + '_psds.hdf'))
            filenames.append(filename)
            with h5py.File(filename, 'w') as psd_file:
                psd_file.attrs['low_frequency_cutoff'] = 10.0
                psd_file.attrs['dynamic_range_factor'] = np.e
                psd_group = psd_file.create_group(ifo)
                psd_group['start_time'] = (np.arange(5) - 0.5) * np.pi**3
                psd_group['end_time'] = (np.arange(5) + 0.5) * np.pi**3
                psd_group = psd_group.create_group('psds')
                for j in range(5):
                    psd_group[str(j)] = np.concatenate(
                        (np.zeros(5), np.arange(5)**2)) * np.e**2
                    psd_group[str(j)].attrs['delta_f'] = 2.0

    with pytest.raises(
            ValueError,
            match='You must provide exactly one coinc file.'):
        source = events.open(*filenames[1:])

    with pytest.raises(
            ValueError,
            match='You must provide exactly one template bank file.'):
        source = events.open(*filenames[:1])

    with pytest.raises(
            ValueError,
            match='You must provide PSD files.'):
        source = events.open(*filenames[:2])

    with pytest.raises(
            ValueError,
            match='You must provide trigger files.'):
        source = events.open(*filenames[:2], filenames[3], filenames[5])

    # Test reading from filenames
    source = events.open(*filenames)
    assert len(source) == 5
    for coinc_id, coinc in source.items():
        for i, (ifo, single) in enumerate(zip(ifos, coinc.singles)):
            assert single.detector == ifo
            assert single.snr == i + coinc_id * np.pi
            assert single.phase == i + coinc_id * np.pi**2
            assert single.time == (
                single.zerolag_time +
                coinc_id * 10 * np.e * (-0.5 if i == 0 else +0.5))
            assert single.zerolag_time == i + coinc_id * np.pi**3
            assert single.psd.f0 == 10.0
            assert single.psd.deltaF == 2.0
            assert np.all(single.psd.data.data == np.arange(5)**2)
        assert coinc.template_args == {}

    # Test reading from h5py.File instances
    events.open(*(h5py.File(filename, 'r') for filename in filenames))

    # Test reading from file-like objects
    with ExitStack() as stack:
        events.open(*(
            stack.enter_context(open(filename, 'rb'))
            for filename in filenames))



#
# Copyright (C) 2019-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Rough-cut injection tool.

The idea is to efficiently sample events, uniformly in "sensitive volume"
(differential comoving volume divided by 1 + z), and from a distribution of
masses and spins, such that later detection cuts will not reject an excessive
number of events.

This occurs in two steps. First, we divide the intrinsic parameter space into a
very coarse 10x10x10x10 grid and calculate the maximum horizon distance in each
grid cell. Second, we directly sample injections jointly from the mass and spin
distribution and a uniform and isotropic spatial distribution with a redshift
cutoff that is piecewise constant in the masses and spins.
"""

from functools import partial

from astropy import cosmology
from astropy.cosmology._utils import vectorize_redshift_method
from astropy import units
from astropy.units import dimensionless_unscaled
import numpy as np
from scipy.integrate import quad, fixed_quad
from scipy.interpolate import interp1d
from scipy.optimize import root_scalar
from scipy.ndimage import maximum_filter

from ..util import progress_map
from ..bayestar.filter import sngl_inspiral_psd
from . import (
    ArgumentParser, FileType, get_random_parser, register_to_xmldoc,
    write_fileobj)


def get_decisive_snr(snrs, min_triggers):
    """Return the SNR for the trigger that decides if an event is detectable.

    Parameters
    ----------
    snrs : list
        List of SNRs (floats).
    min_triggers : int
        Minimum number of triggers to form a coincidence.

    Returns
    -------
    decisive_snr : float

    """
    return sorted(snrs)[-min_triggers]


def lo_hi_nonzero(x):
    nonzero = np.flatnonzero(x)
    return nonzero[0], nonzero[-1]


class GWCosmo:
    """Evaluate GW distance figures of merit for a given cosmology.

    Parameters
    ----------
    cosmo : :class:`astropy.cosmology.FLRW`
        The cosmological model.

    """

    def __init__(self, cosmology):
        self.cosmo = cosmology

    def z_at_snr(self, psds, waveform, f_low, snr_threshold, min_triggers,
                 mass1, mass2, spin1z, spin2z):
        """
        Get redshift at which a waveform attains a given SNR.

        Parameters
        ----------
        psds : list
            List of :class:`lal.REAL8FrequencySeries` objects.
        waveform : str
            Waveform approximant name.
        f_low : float
            Low-frequency cutoff for template.
        snr_threshold : float
            Minimum single-detector SNR.
        min_triggers : int
            Minimum number of triggers to form a coincidence.
        params : list
            List of waveform parameters: mass1, mass2, spin1z, spin2z.

        Returns
        -------
        comoving_distance : float
            Comoving distance in Mpc.

        """
        # Construct waveform
        series = sngl_inspiral_psd(waveform, f_low=f_low,
                                   mass1=mass1, mass2=mass2,
                                   spin1z=spin1z, spin2z=spin2z)
        i_lo, i_hi = lo_hi_nonzero(series.data.data)
        log_f = np.log(series.f0 + series.deltaF * np.arange(i_lo, i_hi + 1))
        log_f_lo = log_f[0]
        log_f_hi = log_f[-1]
        num = interp1d(
            log_f, np.log(series.data.data[i_lo:i_hi + 1]),
            fill_value=-np.inf, bounds_error=False, assume_sorted=True)

        denoms = []
        for series in psds:
            i_lo, i_hi = lo_hi_nonzero(
                np.isfinite(series.data.data) & (series.data.data != 0))
            log_f = np.log(
                series.f0 + series.deltaF * np.arange(i_lo, i_hi + 1))
            denom = interp1d(
                log_f, log_f - np.log(series.data.data[i_lo:i_hi + 1]),
                fill_value=-np.inf, bounds_error=False, assume_sorted=True)
            denoms.append(denom)

        def snr_at_z(z):
            logzp1 = np.log(z + 1)
            integrand = lambda log_f: [
                np.exp(num(log_f + logzp1) + denom(log_f)) for denom in denoms]
            integrals, _ = fixed_quad(
                integrand, log_f_lo, log_f_hi - logzp1, n=1024)
            snr = get_decisive_snr(np.sqrt(4 * integrals), min_triggers)
            with np.errstate(divide='ignore'):
                snr /= self.cosmo.angular_diameter_distance(z).to_value(
                    units.Mpc)
            return snr

        def root_func(z):
            return snr_at_z(z) - snr_threshold

        return root_scalar(root_func, bracket=(0, 1e3)).root

    def get_max_z(self, psds, waveform, f_low, snr_threshold, min_triggers,
                  mass1, mass2, spin1z, spin2z, jobs=1):
        # Calculate the maximum distance on the grid.
        params = [mass1, mass2, spin1z, spin2z]
        shape = np.broadcast_shapes(*(param.shape for param in params))
        result = list(progress_map(
            partial(self.z_at_snr, psds, waveform, f_low,
                    snr_threshold, min_triggers),
            *(param.ravel() for param in params),
            jobs=jobs))
        result = np.reshape(result, shape)

        assert np.all(result >= 0), 'some redshifts are negative'
        assert np.all(np.isfinite(result)), 'some redshifts are not finite'
        return result

    @vectorize_redshift_method
    def _sensitive_volume_integral(self, z):
        dh3_sr = self.cosmo.hubble_distance**3 / units.sr

        def integrand(z):
            result = self.cosmo.differential_comoving_volume(z)
            result /= (1 + z) * dh3_sr
            return result.to_value(dimensionless_unscaled)

        result, _ = quad(integrand, 0, z)
        return result

    def sensitive_volume(self, z):
        """Sensitive volume :math:`V(z)` out to redshift :math:`z`.

        Given a population of events that occur at a constant rate density
        :math:`R` per unit comoving volume per unit proper time, the number of
        observed events out to a redshift :math:`N(z)` over an observation time
        :math:`T` is :math:`N(z) = R T V(z)`.
        """
        dh3 = self.cosmo.hubble_distance**3
        return 4 * np.pi * dh3 * self._sensitive_volume_integral(z)

    def sensitive_distance(self, z):
        r"""Sensitive distance as a function of redshift :math:`z`.

        The sensitive distance is the distance :math:`d_s(z)` defined such that
        :math:`V(z) = 4/3\pi {d_s(z)}^3`, where :math:`V(z)` is the sensitive
        volume.
        """
        dh = self.cosmo.hubble_distance
        return dh * np.cbrt(3 * self._sensitive_volume_integral(z))


def cell_max(values):
    r"""
    Find pairwise max of consecutive elements across all axes of an array.

    Parameters
    ----------
    values : :class:`numpy.ndarray`
        An input array of :math:`n` dimensions,
        :math:`(m_0, m_1, \dots, m_{n-1})`.

    Returns
    -------
    maxima : :class:`numpy.ndarray`
        An input array of :math:`n` dimensions, each with a length 1 less than
        the input array,
        :math:`(m_0 - 1, m_1 - 1, \dots, m_{n-1} - 1)`.

    """
    maxima = maximum_filter(values, size=2, mode='constant')
    indices = (slice(1, None),) * np.ndim(values)
    return maxima[indices]


def assert_not_reached():  # pragma: no cover
    raise AssertionError('This line should not be reached.')


def parser():
    parser = ArgumentParser(parents=[get_random_parser()])
    parser.add_argument(
        '--cosmology', choices=cosmology.available,
        default='Planck15', help='Cosmological model')

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--distribution', help='Use a preset distribution', choices=(
            'bns_astro', 'bns_broad', 'nsbh_astro', 'nsbh_broad',
            'bbh_astro', 'bbh_broad'))
    group.add_argument(
        '--distribution-samples',
        help='Load samples of the intrinsic mass and spin distribution from '
             'any file that can be read as an Astropy table. The table '
             'columns should be mass1, mass2, spin1z, and spin2z.')

    parser.add_argument(
        '--reference-psd', type=FileType('rb'), metavar='PSD.xml[.gz]',
        required=True, help='PSD file')
    parser.add_argument(
        '--f-low', type=float, default=25.0,
        help='Low frequency cutoff in Hz')
    parser.add_argument(
        '--snr-threshold', type=float, default=4.,
        help='Single-detector SNR threshold')
    parser.add_argument(
        '--min-triggers', type=int, default=2,
        help='Emit coincidences only when at least this many triggers '
        'are found')
    parser.add_argument(
        '--min-snr', type=float,
        help='Minimum decisive SNR of injections given the reference PSDs. '
             'Deprecated; use the synonymous --snr-threshold option instead.')
    parser.add_argument(
        '--max-distance', type=float, metavar='Mpc',
        help='Maximum luminosity distance for injections')
    parser.add_argument(
        '--waveform', default='o2-uberbank',
        help='Waveform approximant')
    parser.add_argument(
        '--nsamples', type=int, default=100000,
        help='Output this many injections')
    parser.add_argument(
        '-o', '--output', type=FileType('wb'), default='-',
        metavar='INJ.xml[.gz]', help='Output file, optionally gzip-compressed')
    parser.add_argument(
        '-j', '--jobs', type=int, default=1, const=None, nargs='?',
        help='Number of threads')
    return parser


def main(args=None):
    p = parser()
    with p.parse_args(args) as args:
        import warnings

        from astropy.table import Table
        from ligo.lw import lsctables
        from ligo.lw import utils as ligolw_utils
        from ligo.lw import ligolw
        import lal.series
        from scipy import stats

        if args.min_snr is not None:
            warnings.warn(
                'The --min-snr threshold option is deprecated. '
                'Please use the synonymous --snr-threshold option instead.',
                UserWarning)
            args.snr_threshold = args.min_snr

        xmldoc = ligolw.Document()
        xmlroot = xmldoc.appendChild(ligolw.LIGO_LW())
        process = register_to_xmldoc(xmldoc, p, args)

        # Read PSDs
        psds = list(
            lal.series.read_psd_xmldoc(
                ligolw_utils.load_fileobj(
                    args.reference_psd,
                    contenthandler=lal.series.PSDContentHandler)).values())

        if len(psds) < args.min_triggers:
            parser.error(
                f'The number of PSDs ({len(psds)}) must be greater than or '
                f'equal to the value of --min-triggers ({args.min_triggers}).')

        gwcosmo = GWCosmo(getattr(cosmology, args.cosmology))

        if args.distribution:
            ns_mass_min = 1.0
            ns_mass_max = 2.0
            bh_mass_min = 5.0
            bh_mass_max = 50.0

            ns_astro_spin_min = -0.05
            ns_astro_spin_max = +0.05
            ns_astro_mass_dist = stats.norm(1.33, 0.09)
            ns_astro_spin_dist = stats.uniform(
                ns_astro_spin_min, ns_astro_spin_max - ns_astro_spin_min)

            ns_broad_spin_min = -0.4
            ns_broad_spin_max = +0.4
            ns_broad_mass_dist = stats.uniform(
                ns_mass_min, ns_mass_max - ns_mass_min)
            ns_broad_spin_dist = stats.uniform(
                ns_broad_spin_min, ns_broad_spin_max - ns_broad_spin_min)

            bh_astro_spin_min = -0.99
            bh_astro_spin_max = +0.99
            bh_astro_mass_dist = stats.pareto(b=1.3)
            bh_astro_spin_dist = stats.uniform(
                bh_astro_spin_min, bh_astro_spin_max - bh_astro_spin_min)

            bh_broad_spin_min = -0.99
            bh_broad_spin_max = +0.99
            bh_broad_mass_dist = stats.reciprocal(bh_mass_min, bh_mass_max)
            bh_broad_spin_dist = stats.uniform(
                bh_broad_spin_min, bh_broad_spin_max - bh_broad_spin_min)

            if args.distribution.startswith('bns_'):
                m1_min = m2_min = ns_mass_min
                m1_max = m2_max = ns_mass_max
                if args.distribution.endswith('_astro'):
                    x1_min = x2_min = ns_astro_spin_min
                    x1_max = x2_max = ns_astro_spin_max
                    m1_dist = m2_dist = ns_astro_mass_dist
                    x1_dist = x2_dist = ns_astro_spin_dist
                elif args.distribution.endswith('_broad'):
                    x1_min = x2_min = ns_broad_spin_min
                    x1_max = x2_max = ns_broad_spin_max
                    m1_dist = m2_dist = ns_broad_mass_dist
                    x1_dist = x2_dist = ns_broad_spin_dist
                else:  # pragma: no cover
                    assert_not_reached()
            elif args.distribution.startswith('nsbh_'):
                m1_min = bh_mass_min
                m1_max = bh_mass_max
                m2_min = ns_mass_min
                m2_max = ns_mass_max
                if args.distribution.endswith('_astro'):
                    x1_min = bh_astro_spin_min
                    x1_max = bh_astro_spin_max
                    x2_min = ns_astro_spin_min
                    x2_max = ns_astro_spin_max
                    m1_dist = bh_astro_mass_dist
                    m2_dist = ns_astro_mass_dist
                    x1_dist = bh_astro_spin_dist
                    x2_dist = ns_astro_spin_dist
                elif args.distribution.endswith('_broad'):
                    x1_min = bh_broad_spin_min
                    x1_max = bh_broad_spin_max
                    x2_min = ns_broad_spin_min
                    x2_max = ns_broad_spin_max
                    m1_dist = bh_broad_mass_dist
                    m2_dist = ns_broad_mass_dist
                    x1_dist = bh_broad_spin_dist
                    x2_dist = ns_broad_spin_dist
                else:  # pragma: no cover
                    assert_not_reached()
            elif args.distribution.startswith('bbh_'):
                m1_min = m2_min = bh_mass_min
                m1_max = m2_max = bh_mass_max
                if args.distribution.endswith('_astro'):
                    x1_min = x2_min = bh_astro_spin_min
                    x1_max = x2_max = bh_astro_spin_max
                    m1_dist = m2_dist = bh_astro_mass_dist
                    x1_dist = x2_dist = bh_astro_spin_dist
                elif args.distribution.endswith('_broad'):
                    x1_min = x2_min = bh_broad_spin_min
                    x1_max = x2_max = bh_broad_spin_max
                    m1_dist = m2_dist = bh_broad_mass_dist
                    x1_dist = x2_dist = bh_broad_spin_dist
                else:  # pragma: no cover
                    assert_not_reached()
            else:  # pragma: no cover
                assert_not_reached()

            dists = (m1_dist, m2_dist, x1_dist, x2_dist)

            # Construct mass1, mass2, spin1z, spin2z grid.
            m1 = np.geomspace(m1_min, m1_max, 10)
            m2 = np.geomspace(m2_min, m2_max, 10)
            x1 = np.linspace(x1_min, x1_max, 10)
            x2 = np.linspace(x2_min, x2_max, 10)
            params = m1, m2, x1, x2

            # Calculate the maximum distance on the grid.
            max_z = gwcosmo.get_max_z(
                psds, args.waveform, args.f_low,
                args.snr_threshold, args.min_triggers,
                *np.meshgrid(m1, m2, x1, x2, indexing='ij'), jobs=args.jobs)
            if args.max_distance is not None:
                new_max_z = cosmology.z_at_value(
                    gwcosmo.cosmo.luminosity_distance,
                    args.max_distance * units.Mpc)
                max_z[max_z > new_max_z] = new_max_z
            max_distance = gwcosmo.sensitive_distance(
                max_z).to_value(units.Mpc)

            # Find piecewise constant approximate upper bound on distance.
            max_distance = cell_max(max_distance)

            # Calculate V * T in each grid cell
            cdfs = [dist.cdf(param) for param, dist in zip(params, dists)]
            cdf_los = [cdf[:-1] for cdf in cdfs]
            cdfs = [np.diff(cdf) for cdf in cdfs]
            probs = np.prod(np.meshgrid(*cdfs, indexing='ij'), axis=0)
            probs /= probs.sum()
            probs *= 4/3*np.pi*max_distance**3
            volume = probs.sum()
            probs /= volume
            probs = probs.ravel()

            # Draw random grid cells
            dist = stats.rv_discrete(values=(np.arange(len(probs)), probs))
            indices = np.unravel_index(
                dist.rvs(size=args.nsamples), max_distance.shape)

            # Draw random intrinsic params from each cell
            cols = {}
            cols['mass1'], cols['mass2'], cols['spin1z'], cols['spin2z'] = [
                dist.ppf(
                    stats.uniform(cdf_lo[i], cdf[i]).rvs(size=args.nsamples))
                for i, dist, cdf_lo, cdf in zip(indices, dists, cdf_los, cdfs)]
        elif args.distribution_samples:
            # Load distribution samples.
            samples = Table.read(args.distribution_samples)

            # Calculate the maximum sensitive distance for each sample.
            max_z = gwcosmo.get_max_z(
                psds, args.waveform, args.f_low,
                args.snr_threshold, args.min_triggers,
                samples['mass1'], samples['mass2'],
                samples['spin1z'], samples['spin2z'], jobs=args.jobs)
            if args.max_distance is not None:
                new_max_z = cosmology.z_at_value(
                    gwcosmo.cosmo.luminosity_distance,
                    args.max_distance * units.Mpc)
                max_z[max_z > new_max_z] = new_max_z
            max_distance = gwcosmo.sensitive_distance(
                max_z).to_value(units.Mpc)

            # Calculate V * T for each sample.
            probs = 1 / len(max_distance)
            probs *= 4/3*np.pi*max_distance**3
            volume = probs.sum()
            probs /= volume

            # Draw weighted samples for the simulated events.
            dist = stats.rv_discrete(values=(np.arange(len(probs)), probs))
            # Note that we do this in small batches because
            # stats.rv_discrete.rvs has quadratic memory usage, number of
            # values times number of samples, which might cause us to run out
            # of memory if we did it all at once.
            n_batches = max(args.nsamples * len(probs) // 1_000_000_000, 1)
            batch_sizes = [len(subarray) for subarray in
                           np.array_split(np.empty(args.nsamples), n_batches)]
            indices = np.concatenate([dist.rvs(size=batch_size)
                                      for batch_size in batch_sizes])

            cols = {key: samples[key][indices]
                    for key in ['mass1', 'mass2', 'spin1z', 'spin2z']}
        else:
            assert_not_reached()

        volumetric_rate = (
            args.nsamples / volume * units.year**-1 * units.Mpc**-3)

        # Swap binary components as needed to ensure that mass1 >= mass2.
        # Note that the .copy() is important.
        # See https://github.com/numpy/numpy/issues/14428
        swap = cols['mass1'] < cols['mass2']
        cols['mass1'][swap], cols['mass2'][swap] = \
            cols['mass2'][swap].copy(), cols['mass1'][swap].copy()
        cols['spin1z'][swap], cols['spin2z'][swap] = \
            cols['spin2z'][swap].copy(), cols['spin1z'][swap].copy()

        # Draw random extrinsic parameters
        cols['distance'] = stats.powerlaw(
            a=3, scale=max_distance[indices]).rvs(size=args.nsamples)
        cols['longitude'] = stats.uniform(0, 2 * np.pi).rvs(
            size=args.nsamples)
        cols['latitude'] = np.arcsin(stats.uniform(-1, 2).rvs(
            size=args.nsamples))
        cols['inclination'] = np.arccos(stats.uniform(-1, 2).rvs(
            size=args.nsamples))
        cols['polarization'] = stats.uniform(0, 2 * np.pi).rvs(
            size=args.nsamples)
        cols['coa_phase'] = stats.uniform(-np.pi, 2 * np.pi).rvs(
            size=args.nsamples)
        cols['time_geocent'] = stats.uniform(
            1e9, units.year.to(units.second)).rvs(size=args.nsamples)

        # Convert from sensitive distance to redshift and comoving distance.
        # FIXME: Replace this brute-force lookup table with a solver.
        z = np.linspace(0, max_z.max(), 10000)
        ds = gwcosmo.sensitive_distance(z).to_value(units.Mpc)
        dc = gwcosmo.cosmo.comoving_distance(z).to_value(units.Mpc)
        z_for_ds = interp1d(ds, z, kind='cubic', assume_sorted=True)
        dc_for_ds = interp1d(ds, dc, kind='cubic', assume_sorted=True)
        zp1 = 1 + z_for_ds(cols['distance'])
        cols['distance'] = dc_for_ds(cols['distance'])

        # Apply redshift factor to convert from comoving distance and source
        # frame masses to luminosity distance and observer frame masses.
        for key in ['distance', 'mass1', 'mass2']:
            cols[key] *= zp1

        # Populate sim_inspiral table
        sims = xmlroot.appendChild(lsctables.New(lsctables.SimInspiralTable))
        for row in zip(*cols.values()):
            sims.appendRow(
                **dict(
                    dict.fromkeys(sims.validcolumns, None),
                    process_id=process.process_id,
                    simulation_id=sims.get_next_id(),
                    waveform=args.waveform,
                    f_lower=args.f_low,
                    **dict(zip(cols.keys(), row))))

        # Record process end time.
        process.comment = str(volumetric_rate)
        process.set_end_time_now()

        # Write output file.
        write_fileobj(xmldoc, args.output)



#
# Copyright (C) 2017-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Support for HEALPix UNIQ pixel indexing [1]_ and multi-order coverage (MOC)
maps [2]_.

References
----------
.. [1] Reinecke & Hivon, 2015. "Efficient data structures for masks on 2D
       grids." AA 580, A132. :doi:`10.1051/0004-6361/201526549`
.. [2] Boch et al., 2014. "MOC - HEALPix Multi-Order Coverage map." IVOA
       Recommendation <http://ivoa.net/documents/MOC/>.

"""

from astropy import table
from astropy import units as u
import astropy_healpix as ah
import numpy as np
from numpy.lib.recfunctions import repack_fields

from .core import nest2uniq, uniq2nest, uniq2order, uniq2pixarea, uniq2ang
from .core import rasterize as _rasterize
from .util.numpy import add_newdoc_ufunc

__all__ = ('nest2uniq', 'uniq2nest', 'uniq2order', 'uniq2pixarea',
           'uniq2ang', 'rasterize', 'bayestar_adaptive_grid')


add_newdoc_ufunc(nest2uniq, """\
Convert a pixel index from NESTED to NUNIQ ordering.

Parameters
----------
order : `numpy.ndarray`
    HEALPix resolution order, the logarithm base 2 of `nside`
ipix : `numpy.ndarray`
    NESTED pixel index

Returns
-------
uniq : `numpy.ndarray`
    NUNIQ pixel index

""")


add_newdoc_ufunc(uniq2order, """\
Determine the HEALPix resolution order of a HEALPix NESTED index.

Parameters
----------
uniq : `numpy.ndarray`
    NUNIQ pixel index

Returns
-------
order : `numpy.ndarray`
    HEALPix resolution order, the logarithm base 2 of `nside`

""")


add_newdoc_ufunc(uniq2pixarea, """\
Determine the area of a HEALPix NESTED index.

Parameters
----------
uniq : `numpy.ndarray`
    NUNIQ pixel index

Returns
-------
area : `numpy.ndarray`
    The pixel's area in steradians

""")


add_newdoc_ufunc(uniq2nest, """\
Convert a pixel index from NUNIQ to NESTED ordering.

Parameters
----------
uniq : `numpy.ndarray`
    NUNIQ pixel index

Returns
-------
order : `numpy.ndarray`
    HEALPix resolution order (logarithm base 2 of `nside`)
ipix : `numpy.ndarray`
    NESTED pixel index

""")


def rasterize(moc_data, order=None):
    """Convert a multi-order HEALPix dataset to fixed-order NESTED ordering.

    Parameters
    ----------
    moc_data : `numpy.ndarray`
        A multi-order HEALPix dataset stored as a Numpy record array whose
        first column is called UNIQ and contains the NUNIQ pixel index. Every
        point on the unit sphere must be contained in exactly one pixel in the
        dataset.
    order : int, optional
        The desired output resolution order, or :obj:`None` for the maximum
        resolution present in the dataset.

    Returns
    -------
    nested_data : `numpy.ndarray`
        A fixed-order, NESTED-ordering HEALPix dataset with all of the columns
        that were in moc_data, with the exception of the UNIQ column.

    """
    if np.ndim(moc_data) != 1:
        raise ValueError('expected 1D structured array or Astropy table')
    elif order is None or order < 0:
        order = -1
    else:
        orig_order, orig_nest = uniq2nest(moc_data['UNIQ'])
        to_downsample = order < orig_order
        if np.any(to_downsample):
            to_keep = table.Table(moc_data[~to_downsample], copy=False)
            orig_order = orig_order[to_downsample]
            orig_nest = orig_nest[to_downsample]
            to_downsample = table.Table(moc_data[to_downsample], copy=False)

            ratio = 1 << (2 * np.int64(orig_order - order))
            weights = 1.0 / ratio
            for colname, column in to_downsample.columns.items():
                if colname != 'UNIQ':
                    column *= weights
            to_downsample['UNIQ'] = nest2uniq(
                np.int8(order), orig_nest // ratio)
            to_downsample = to_downsample.group_by(
                'UNIQ').groups.aggregate(np.sum)

            moc_data = table.vstack((to_keep, to_downsample))

    # Ensure that moc_data has appropriate padding for each of its columns to
    # be properly aligned in order to avoid undefined behavior.
    moc_data = repack_fields(np.asarray(moc_data), align=True)

    return _rasterize(moc_data, order=order)


def bayestar_adaptive_grid(probdensity, *args, top_nside=16, rounds=8,
                           **kwargs):
    """Create a sky map by evaluating a function on an adaptive grid.

    Perform the BAYESTAR adaptive mesh refinement scheme as described in
    Section VI of Singer & Price 2016, PRD, 93, 024013
    :doi:`10.1103/PhysRevD.93.024013`. This computes the sky map
    using a provided analytic function and refines the grid, dividing the
    highest 25% into subpixels and then recalculating their values. The extra
    given args and kwargs will be passed to the given probdensity function.

    Parameters
    ----------
    probdensity : callable
        Probability density function. The first argument consists of
        column-stacked array of right ascension and declination in radians.
        The return value must be a 1D array of the probability density in
        inverse steradians with the same length as the argument.
    top_nside : int
        HEALPix NSIDE resolution of initial evaluation of the sky map
    rounds : int
        Number of refinement rounds, including the initial sky map evaluation

    Returns
    -------
    skymap : astropy.table.Table
        An astropy Table with UNIQ and PROBDENSITY columns, representing
        a multi-ordered sky map
    """
    top_npix = ah.nside_to_npix(top_nside)
    nrefine = top_npix // 4
    cells = zip([0] * nrefine, [top_nside // 2] * nrefine, range(nrefine))
    for iround in range(rounds + 1):
        print('adaptive refinement round {} of {} ...'.format(
            iround, rounds))
        cells = sorted(cells, key=lambda p_n_i: p_n_i[0] / p_n_i[1]**2)
        new_nside, new_ipix = np.transpose([
            (nside * 2, ipix * 4 + i)
            for _, nside, ipix in cells[-nrefine:] for i in range(4)])
        ra, dec = ah.healpix_to_lonlat(new_ipix, new_nside, order='nested')
        p = probdensity(np.column_stack((ra.value, dec.value)),
                        *args, **kwargs)
        cells[-nrefine:] = zip(p, new_nside, new_ipix)

    """Return a HEALPix multi-order map of the posterior density."""
    post, nside, ipix = zip(*cells)
    post = np.asarray(list(post))
    nside = np.asarray(list(nside))
    ipix = np.asarray(list(ipix))

    # Make sure that sky map is normalized (it should be already)
    post /= np.sum(post * ah.nside_to_pixel_area(nside).to_value(u.sr))

    # Convert from NESTED to UNIQ pixel indices
    order = np.log2(nside).astype(int)
    uniq = nest2uniq(order.astype(np.int8), ipix)

    # Done!
    return table.Table([uniq, post], names=['UNIQ', 'PROBDENSITY'],
                       copy=False)


del add_newdoc_ufunc



import os
import pkgutil

__all__ = ()

# Import all symbols from all submodules of this module.
for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
    if module not in {'tests'}:
        exec('from . import {0};'
             '__all__ += getattr({0}, "__all__", ());'
             'from .{0} import *'.format(module))
    del module

# Clean up
del os, pkgutil



#
# Copyright (C) 2013-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Markov-Chain Monte Carlo sky localization."""

from . import (
    ArgumentParser, FileType, get_waveform_parser,
    get_posterior_parser, get_random_parser)


def parser():
    parser = ArgumentParser(parents=[get_waveform_parser(),
                                     get_posterior_parser(),
                                     get_random_parser()])
    parser.add_argument(
        'input', metavar='INPUT.{hdf,xml,xml.gz,sqlite}', default='-',
        nargs='+', type=FileType('rb'),
        help='Input LIGO-LW XML file, SQLite file, or PyCBC HDF5 files. '
             'For PyCBC, you must supply the coincidence file '
             '(e.g. "H1L1-HDFINJFIND.hdf" or "H1L1-STATMAP.hdf"), '
             'the template bank file (e.g. H1L1-BANK2HDF.hdf), '
             'the single-detector merged PSD files '
             '(e.g. "H1-MERGE_PSDS.hdf" and "L1-MERGE_PSDS.hdf"), '
             'and the single-detector merged trigger files '
             '(e.g. "H1-HDF_TRIGGER_MERGE.hdf" and '
             '"L1-HDF_TRIGGER_MERGE.hdf"), '
             'in any order.')
    parser.add_argument(
        '--pycbc-sample', default='foreground',
        help='(PyCBC only) sample population')
    parser.add_argument(
        '--coinc-event-id', type=int, nargs='*',
        help='run on only these specified events')
    parser.add_argument(
        '--output', '-o', default='.',
        help='output directory')
    parser.add_argument(
        '--condor-submit', action='store_true',
        help='submit to Condor instead of running locally')

    group = parser.add_argument_group(
        'fixed parameter options',
        'Options to hold certain parameters constant')
    group.add_argument('--ra', type=float, metavar='DEG',
                       help='Right ascension')
    group.add_argument('--dec', type=float, metavar='DEG',
                       help='Declination')
    group.add_argument('--distance', type=float, metavar='Mpc',
                       help='Luminosity distance')

    return parser


def identity(x):
    return x


def main(args=None):
    with parser().parse_args(args) as opts:
        import logging
        log = logging.getLogger('BAYESTAR')

        # BAYESTAR imports.
        from ..io import events, hdf5
        from ..bayestar import condition, condition_prior, ez_emcee, log_post

        # Other imports.
        from astropy.table import Table
        import numpy as np
        import os
        import subprocess
        import sys

        import lal

        # Read coinc file.
        log.info(
            '%s:reading input files', ','.join(
                file.name for file in opts.input))
        event_source = events.open(*opts.input, sample=opts.pycbc_sample)

        os.makedirs(opts.output, exist_ok=True)

        if opts.condor_submit:
            if opts.seed is not None:
                raise NotImplementedError(
                    '--seed does not yet work with --condor-submit')
            if opts.coinc_event_id:
                raise ValueError(
                    'must not set --coinc-event-id with --condor-submit')
            with subprocess.Popen(['condor_submit'],
                                  text=True, stdin=subprocess.PIPE) as proc:
                f = proc.stdin
                print('''
                    accounting_group = ligo.dev.o4.cbc.pe.bayestar
                    on_exit_remove = (ExitBySignal == False) && (ExitCode == 0)
                    on_exit_hold = (ExitBySignal == True) || (ExitCode != 0)
                    on_exit_hold_reason = (ExitBySignal == True \
                        ? strcat("The job exited with signal ", ExitSignal) \
                        : strcat("The job exited with code ", ExitCode))
                    request_memory = 1000 MB
                    universe = vanilla
                    getenv = true
                    executable = /usr/bin/env
                    JobBatchName = BAYESTAR
                    environment = "OMP_NUM_THREADS=1"
                    ''', file=f)
                print(
                    'error =', os.path.join(opts.output, '$(cid).err'), file=f)
                print('log =', os.path.join(opts.output, '$(cid).log'), file=f)
                print(
                    'arguments = "',
                    *(arg for arg in sys.argv if arg != '--condor-submit'),
                    '--coinc-event-id $(cid)"', file=f)
                print('queue cid in', *event_source, file=f)
            sys.exit(proc.returncode)

        if opts.coinc_event_id:
            event_source = {
                key: event_source[key] for key in opts.coinc_event_id}

        # Loop over all sngl_inspiral <-> sngl_inspiral coincs.
        for int_coinc_event_id, event in event_source.items():
            coinc_event_id = 'coinc_event:coinc_event_id:{}'.format(
                int_coinc_event_id)

            log.info('%s:preparing', coinc_event_id)

            epoch, sample_rate, epochs, snrs, responses, locations, horizons \
                = condition(event, waveform=opts.waveform, f_low=opts.f_low,
                            enable_snr_series=opts.enable_snr_series,
                            f_high_truncate=opts.f_high_truncate)

            min_distance, max_distance, prior_distance_power, cosmology \
                = condition_prior(horizons, opts.min_distance,
                                  opts.max_distance, opts.prior_distance_power,
                                  opts.cosmology)

            gmst = lal.GreenwichMeanSiderealTime(epoch)

            max_abs_t = 2 * snrs.data.shape[1] / sample_rate
            xmin = [0, -1, min_distance, -1, 0, 0]
            xmax = [2 * np.pi, 1, max_distance, 1, 2 * np.pi, 2 * max_abs_t]
            names = 'ra dec distance inclination twopsi time'.split()
            transformed_names = 'ra sin_dec distance u twopsi time'.split()
            forward_transforms = [identity, np.sin, identity,
                                  np.cos, identity, identity]
            reverse_transforms = [identity, np.arcsin, identity,
                                  np.arccos, identity, identity]
            kwargs = dict(min_distance=min_distance, max_distance=max_distance,
                          prior_distance_power=prior_distance_power,
                          cosmology=cosmology, gmst=gmst,
                          sample_rate=sample_rate, epochs=epochs, snrs=snrs,
                          responses=responses, locations=locations,
                          horizons=horizons)

            # Fix parameters
            for i, key in reversed(list(enumerate(['ra', 'dec', 'distance']))):
                value = getattr(opts, key)
                if value is None:
                    continue

                if key in ['ra', 'dec']:
                    # FIXME: figure out a more elegant way to address different
                    # units in command line arguments and posterior samples
                    value = np.deg2rad(value)

                kwargs[transformed_names[i]] = forward_transforms[i](value)
                del (xmin[i], xmax[i], names[i], transformed_names[i],
                     forward_transforms[i], reverse_transforms[i])

            log.info('%s:sampling', coinc_event_id)

            # Run MCMC
            chain = ez_emcee(
                log_post, xmin, xmax, kwargs=kwargs, vectorize=True)

            # Transform back from sin_dec to dec and cos_inclination to
            # inclination
            for i, func in enumerate(reverse_transforms):
                chain[:, i] = func(chain[:, i])

            # Create Astropy table
            chain = Table(rows=chain, names=names, copy=False)

            log.info('%s:saving posterior samples', coinc_event_id)

            hdf5.write_samples(
                chain,
                os.path.join(
                    opts.output, '{}.hdf5'.format(int_coinc_event_id)),
                path='/bayestar/posterior_samples', overwrite=True)



from astropy.io.fits import Header
from astropy.table import Table
from astropy_healpix import HEALPix
import numpy as np
from reproject import reproject_from_healpix

from ...moc import nest2uniq
from ..reproject_from_healpix_moc import reproject_from_healpix_moc


def test_reproject_from_healpix_moc():
    header = Header({
        'NAXIS': 2,
        'NAXIS1': 180,
        'NAXIS2': 180,
        'CRPIX1': 90.5,
        'CRPIX2': 90.5,
        'CRVAL1': 180,
        'CRVAL2': 0,
        'CDELT1': 1,
        'CDELT2': 1,
        'CTYPE1': 'RA---CAR',
        'CTYPE2': 'DEC--CAR',
        'RADESYS': 'ICRS',
    })
    hpx = HEALPix(nside=32)
    map_flat = np.arange(hpx.npix)

    map_moc = Table({
        'UNIQ': nest2uniq(np.int8(hpx.level), hpx.ring_to_nested(map_flat)),
        'IMG': map_flat,
    })

    expected = reproject_from_healpix(
        (map_flat, 'ICRS'), header, order='nearest-neighbor', nested=False)
    result = reproject_from_healpix_moc(
        (map_moc, 'ICRS'), header, 'nearest-neighbor')

    np.testing.assert_array_equal(result, expected)



#
# Copyright (C) 2011-2024  Will M. Farr <will.farr@ligo.org>
#                          Leo P. Singer <leo.singer@ligo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Generate a FITS sky map file from posterior samples using clustering and
kernel density estimation.

The input file should be an HDF5 file with the following columns:

*  ``ra``, ``rightascension``, or ``right_ascension``: J2000 right ascension in
    radians
*  ``dec`` or ``declination``: J200 declination in radians
*  ``dist``, ``distance``, or ``luminosity_distance``: luminosity distance in
   Mpc (optional)

The output consist of two files:

*  ``skypost.obj``, a :mod:`pickle` representation of the kernel density
   estimator
*  ``skymap.fits.gz``, a 3D localization in HEALPix/FITS format
"""

from argparse import SUPPRESS

from . import (
    ArgumentParser, DirType, EnableAction, FileType, get_random_parser)


def parser():
    # Command line interface.
    parser = ArgumentParser(parents=[get_random_parser()])
    parser.add_argument('samples', type=FileType('rb'), metavar='SAMPLES.hdf5',
                        help='posterior samples file')
    # Only present for backward compatibility with --samples syntax
    parser.add_argument('--samples', action='store_false', dest='_ignored',
                        help=SUPPRESS)
    parser.add_argument('--outdir', '-o', default='.',
                        type=DirType(create=True), help='output directory')
    parser.add_argument('--fitsoutname', default='skymap.fits',
                        metavar='SKYMAP.fits[.gz]',
                        help='filename for the FITS file')
    parser.add_argument('--loadpost', type=FileType('rb'),
                        metavar='SKYPOST.obj',
                        help='filename for pickled posterior state')
    parser.add_argument('--maxpts', type=int,
                        help='maximum number of posterior points to use; if '
                        'omitted or greater than or equal to the number of '
                        'posterior samples, then use all samples')
    parser.add_argument('--trials', type=int, default=5,
                        help='number of trials at each clustering number')
    parser.add_argument('--enable-distance-map', action=EnableAction,
                        help='generate HEALPix map of distance estimates')
    parser.add_argument('--enable-multiresolution', action=EnableAction,
                        default=True,
                        help='generate a multiresolution HEALPix map')
    parser.add_argument('--top-nside', type=int, default=16,
                        help='choose a start nside before HEALPix refinement '
                        'steps (must be a valid nside)')
    parser.add_argument('-j', '--jobs', type=int, default=1, const=None,
                        nargs='?', help='Number of threads')
    parser.add_argument('--instruments', metavar='H1|L1|V1|...', nargs='+',
                        help='instruments to store in FITS header')
    parser.add_argument('--objid', help='event ID to store in FITS header')
    parser.add_argument('--path', type=str, default=None,
                        help="The path of the dataset within the HDF5 file")
    parser.add_argument('--tablename', type=str, default='posterior_samples',
                        help='The name of the table to search for recursively '
                        'within the HDF5 file. By default, search for '
                        'posterior_samples')
    return parser


def main(args=None):
    _parser = parser()
    with _parser.parse_args(args) as args:
        # Late imports
        from .. import io
        from ..io.hdf5 import _remap_colnames
        from ..bayestar import rasterize
        from .. import version
        from astropy.table import Table
        from astropy.time import Time
        import numpy as np
        import os
        import sys
        import pickle
        from ..kde import Clustered2Plus1DSkyKDE, Clustered2DSkyKDE
        import logging
        from textwrap import wrap

        log = logging.getLogger()

        log.info('reading samples')
        try:
            data = io.read_samples(args.samples.name, path=args.path,
                                   tablename=args.tablename)
        except IOError:
            # FIXME: remove this code path once we support only HDF5
            data = Table.read(args.samples, format='ascii')
            _remap_colnames(data)

        if args.maxpts is not None and args.maxpts < len(data):
            log.info('taking random subsample of chain')
            data = data[
                np.random.choice(len(data), args.maxpts, replace=False)]
        try:
            dist = data['dist']
        except KeyError:
            try:
                dist = data['distance']
            except KeyError:
                dist = None

        if args.loadpost is None:
            if dist is None:
                if args.enable_distance_map:
                    _parser.error("The posterior samples file '{}' does not "
                                  "have a distance column named 'dist' or "
                                  "'distance'. Cannot generate distance map. "
                                  "If you do not intend to generate a "
                                  "distance map, then add the "
                                  "'--disable-distance-map' command line "
                                  "argument.".format(args.samples.name))
                pts = np.column_stack((data['ra'], data['dec']))
            else:
                pts = np.column_stack((data['ra'], data['dec'], dist))
            if args.enable_distance_map:
                cls = Clustered2Plus1DSkyKDE
            else:
                cls = Clustered2DSkyKDE
            skypost = cls(pts, trials=args.trials, jobs=args.jobs)

            log.info('pickling')
            with open(os.path.join(args.outdir, 'skypost.obj'), 'wb') as out:
                pickle.dump(skypost, out)
        else:
            skypost = pickle.load(args.loadpost)
            skypost.jobs = args.jobs

        log.info('making skymap')
        hpmap = skypost.as_healpix(top_nside=args.top_nside)
        if not args.enable_multiresolution:
            hpmap = rasterize(hpmap)
        hpmap.meta.update(io.fits.metadata_for_version_module(version))
        hpmap.meta['creator'] = _parser.prog
        hpmap.meta['origin'] = 'LIGO/Virgo/KAGRA'
        hpmap.meta['gps_creation_time'] = Time.now().gps
        hpmap.meta['history'] = [
            '', 'Generated by running the following script:',
            *wrap(' '.join([_parser.prog] + sys.argv[1:]), 72)]
        if args.objid is not None:
            hpmap.meta['objid'] = args.objid
        if args.instruments:
            hpmap.meta['instruments'] = args.instruments
        if args.enable_distance_map:
            hpmap.meta['distmean'] = np.mean(dist)
            hpmap.meta['diststd'] = np.std(dist)

        keys = ['time', 'time_mean', 'time_maxl']
        for key in keys:
            try:
                time = data[key]
            except KeyError:
                continue
            else:
                hpmap.meta['gps_time'] = time.mean()
                break
        else:
            log.warning(
                'Cannot determine the event time from any of the columns %r',
                keys)

        io.write_sky_map(os.path.join(args.outdir, args.fitsoutname),
                         hpmap, nest=True)



from collections import namedtuple
import multiprocessing
import os
import signal

import lal
import pytest

from ...io.events.base import Event, SingleEvent
from .. import localize, rasterize


_MockSingleEvent = namedtuple(
    'MockSingleEvent', 'detector snr phase time psd snr_series')

_MockEvent = namedtuple(
    'MockEvent', 'singles template_args')


class MockSingleEvent(_MockSingleEvent, SingleEvent):
    pass


class MockEvent(_MockEvent, Event):
    pass


template_args = {'mass1': 1.414, 'mass2': 1.414}


@pytest.fixture
def mock_event():
    psd = lal.CreateREAL8FrequencySeries(
        None, 0, 0, 32, lal.DimensionlessUnit, 128)
    psd.data.data[:] = 1
    test_single_event = MockSingleEvent(
        'H1', 12.345, 0.6789, 0.1234, psd, None)
    mock_event = MockEvent([test_single_event], template_args)
    return mock_event


def test_localize_0_detectors():
    """Running on an event with 0 detectors should raise an error."""
    test_event = MockEvent([], template_args)
    with pytest.raises(ValueError):
        localize(test_event)


def test_localize_1_detector(mock_event):
    """Running on an event with 1 detector should produce a sky map that
    reflects the antenna pattern.
    """
    skymap = localize(mock_event)

    # Make sure that none of the extrinsic parameters are in the event history
    history = '\n'.join(skymap.meta['history'])
    for forbidden in ['snr=', '12.345', 'time=', '0.6789', 'phase=', '0.1234',
                      'mass1=', 'mass2=', '1.414']:
        assert forbidden not in history

    # FIXME: work out what this should be
    rasterized = rasterize(skymap)
    assert rasterized


def run_interruptible(event, started, cancelled):
    started.set()
    try:
        localize(event)
    except KeyboardInterrupt:
        cancelled.set()


def test_localize_interruptible(mock_event):
    """Test that localize() stops swiftly and gracefully when interrupted."""
    with multiprocessing.Manager():
        started = multiprocessing.Event()
        cancelled = multiprocessing.Event()
        process = multiprocessing.Process(
            target=run_interruptible, args=(mock_event, started, cancelled))
        try:
            process.start()
            assert started.wait(5)
            os.kill(process.pid, signal.SIGINT)
            assert cancelled.wait(10)
        finally:
            process.join()



# This file is adapted from the Astropy package template, which is licensed
# under a 3-clause BSD style license - see licenses/TEMPLATE_LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
# ----------------------------------------------------------------------------

__all__ = ('omp',)


class Omp:
    """OpenMP runtime settings.

    Attributes
    ----------
    num_threads : int
        Adjust the number of OpenMP threads. Getting and setting this attribute
        call :man:`omp_get_num_threads` and :man:`omp_set_num_threads`
        respectively.

    """

    @property
    def num_threads(self):
        from .core import get_num_threads
        return get_num_threads()

    @num_threads.setter
    def num_threads(self, value):
        from .core import set_num_threads
        set_num_threads(value)


class Itt():
    """Context manager to run code with Intel ITT tracing temporarily enabled.

    If you use this context manager, then you should start profiling with
    collection paused.

    See https://www.intel.com/content/www/us/en/docs/vtune-profiler/user-guide/2023-1/instrumentation-and-tracing-technology-apis.html
    """  # noqa: E501

    def __enter__(self):
        from .core import itt_resume
        itt_resume()
        return self

    def __exit__(self, *args, **kwargs):
        from .core import itt_pause
        itt_pause()


omp = Omp()
del Omp

itt = Itt()
del Itt



#
# Copyright (C) 2017-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Astropy coordinate frame for eigendecomposition of a cloud of points or a 3D
sky map.
"""

from astropy.coordinates import (
    BaseCoordinateFrame, CartesianRepresentation, DynamicMatrixTransform,
    frame_transform_graph, ICRS, SphericalRepresentation)
from astropy.coordinates import CartesianRepresentationAttribute
from astropy.units import dimensionless_unscaled
import numpy as np

from ..distance import principal_axes

__all__ = ('EigenFrame',)


class EigenFrame(BaseCoordinateFrame):
    """A coordinate frame that has its axes aligned with the principal
    components of a cloud of points.
    """

    e_x = CartesianRepresentationAttribute(
        default=CartesianRepresentation(1, 0, 0, unit=dimensionless_unscaled),
        unit=dimensionless_unscaled)
    e_y = CartesianRepresentationAttribute(
        default=CartesianRepresentation(0, 1, 0, unit=dimensionless_unscaled),
        unit=dimensionless_unscaled)
    e_z = CartesianRepresentationAttribute(
        default=CartesianRepresentation(0, 0, 1, unit=dimensionless_unscaled),
        unit=dimensionless_unscaled)

    default_representation = SphericalRepresentation

    @classmethod
    def for_coords(cls, coords):
        """Create a coordinate frame that has its axes aligned with the
        principal components of a cloud of points.

        Parameters
        ----------
        coords : `astropy.coordinates.SkyCoord`
            A cloud of points

        Returns
        -------
        frame : `EigenFrame`
            A new coordinate frame

        """
        v = coords.icrs.cartesian.xyz.value
        _, r = np.linalg.eigh(np.dot(v, v.T))
        r = r[:, ::-1]  # Order by descending eigenvalue
        e_x, e_y, e_z = CartesianRepresentation(r, unit=dimensionless_unscaled)
        return cls(e_x=e_x, e_y=e_y, e_z=e_z)

    @classmethod
    def for_skymap(cls, prob, distmu, distsigma, nest=False):
        """Create a coordinate frame that has its axes aligned with the
        principal components of a 3D sky map.

        Parameters
        ----------
        prob : `numpy.ndarray`
            Marginal probability (pix^-2)
        distmu : `numpy.ndarray`
            Distance location parameter (Mpc)
        distsigma : `numpy.ndarray`
            Distance scale parameter (Mpc)
        distnorm : `numpy.ndarray`
            Distance normalization factor (Mpc^-2)
        nest : bool, default=False
            Indicates whether the input sky map is in nested rather than
            ring-indexed HEALPix coordinates (default: ring).

        Returns
        -------
        frame : `EigenFrame`
            A new coordinate frame

        """
        r = principal_axes(prob, distmu, distsigma, nest=nest)
        r = r[:, ::-1]  # Order by descending eigenvalue
        e_x, e_y, e_z = CartesianRepresentation(r, unit=dimensionless_unscaled)
        return cls(e_x=e_x, e_y=e_y, e_z=e_z)


@frame_transform_graph.transform(DynamicMatrixTransform, ICRS, EigenFrame)
def icrs_to_eigenframe(from_coo, to_frame):
    return np.vstack((to_frame.e_x.xyz.value,
                      to_frame.e_y.xyz.value,
                      to_frame.e_z.xyz.value))


@frame_transform_graph.transform(DynamicMatrixTransform, EigenFrame, ICRS)
def eigenframe_to_icrs(from_coo, to_frame):
    return np.column_stack((from_coo.e_x.xyz.value,
                            from_coo.e_y.xyz.value,
                            from_coo.e_z.xyz.value))



#
# Copyright (C) 2011-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Show a sky map's Bayes factor for coherence vs. incoherence as a bullet chart.
"""

from . import ArgumentParser, FileType
from .matplotlib import get_figure_parser


def parser():
    parser = ArgumentParser(parents=[get_figure_parser()])
    parser.add_argument(
        'input', metavar='INPUT.fits[.gz]', type=FileType('rb'),
        default='-', nargs='?', help='Input FITS file')
    parser.set_defaults(colormap='RdYlBu')
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        # Late imports
        from astropy.io import fits
        import numpy as np
        from ..plot import plot_bayes_factor

        header = fits.getheader(opts.input, 1)
        logb = header['LOGBCI']
        objid = header.get('OBJECT')

        title = 'Coherence'
        if objid:
            title += f' of {objid}'
        logb_string = np.format_float_positional(logb, 1, trim='0', sign=True)
        title += fr' $[\ln\,B = {logb_string}]$'

        plot_bayes_factor(logb, title=title, palette=opts.colormap)

        # Show or save output.
        opts.output()



#
# Copyright (C) 2019-2020  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Tools for progress bars"""

try:
    from billiard import Pool
except ImportError:
    from multiprocessing import Pool
from heapq import heappop, heappush
from operator import length_hint

from tqdm.auto import tqdm

from .. import omp

__all__ = ('progress_map',)


class WrappedFunc:

    def __init__(self, func):
        self.func = func

    def __call__(self, i_args):
        i, args = i_args
        return i, self.func(*args)


def _get_total_estimate(*iterables):
    """Estimate total loop iterations for mapping over multiple iterables."""
    estimates = (length_hint(iterable, -1) for iterable in iterables)
    valid_estimates = (estimate for estimate in estimates if estimate != -1)
    return min(valid_estimates, default=None)


def _results_in_order(completed):
    """Put results back into order and yield them as quickly as they arrive."""
    heap = []
    current = 0
    for i_result in completed:
        i, result = i_result
        if i == current:
            yield result
            current += 1
            while heap and heap[0][0] == current:
                _, result = heappop(heap)
                yield result
                current += 1
        else:
            heappush(heap, i_result)
    assert not heap, 'The heap must be empty'


def _init_process():
    """Disable OpenMP when using multiprocessing."""
    omp.num_threads = 1


def progress_map(func, *iterables, jobs=1, **kwargs):
    """Map a function across iterables of arguments.

    This is comparable to :meth:`astropy.utils.console.ProgressBar.map`, except
    that it is implemented using :mod:`tqdm` and so provides more detailed and
    accurate progress information.
    """
    total = _get_total_estimate(*iterables)
    if jobs == 1:
        yield from tqdm(map(func, *iterables), total=total, **kwargs)
    else:
        with Pool(jobs, _init_process) as pool:
            yield from _results_in_order(
                tqdm(
                    pool.imap_unordered(
                        WrappedFunc(func),
                        enumerate(zip(*iterables))
                    ),
                    total=total, **kwargs
                )
            )



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

import warnings

from astropy.version import version as astropy_version
import pytest

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


def pytest_configure(config):

    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES['astropy'] = 'astropy'
        PYTEST_HEADER_MODULES['astropy-healpix'] = 'astropy_healpix'
        PYTEST_HEADER_MODULES['healpy'] = 'healpy'
        PYTEST_HEADER_MODULES['reproject'] = 'reproject'

        from . import __version__
        packagename = 'ligo.skymap'
        TESTED_VERSIONS[packagename] = __version__


# Uncomment the last two lines in this block to treat all DeprecationWarnings as
# exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
# as follow (although default should work for most cases).
# To ignore some packages that produce deprecation warnings on import
# (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
# 'setuptools'), add:
#     modules_to_ignore_on_import=['module_1', 'module_2']
# To ignore some specific deprecation warning messages for Python version
# MAJOR.MINOR or later, add:
#     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# from astropy.tests.helper import enable_deprecations_as_exceptions  # noqa
# enable_deprecations_as_exceptions()


@pytest.fixture(autouse=True, scope='session')
def treat_unclosed_files_as_errors():
    """Treat warnings abnout unclosed files as errors

    Many of the command-line tools in :mod:`ligo.skymap.tool` use
    :class:`arparse.FileType` and therefore might leave files opened. Treat
    this as an error.

    """
    warnings.filterwarnings('error', 'unclosed file .*', ResourceWarning)
    warnings.filterwarnings(
        'error', category=pytest.PytestUnraisableExceptionWarning)



#
# Copyright (C) 2013-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Synthesize triggers for simulated sources by realizing Gaussian measurement
errors in SNR and time of arrival. The input file (or stdin if the input file
is omitted) should be an optionally gzip-compressed LIGO-LW XML file of the
form produced by lalapps_inspinj. The output file (or stdout if omitted) will
be an optionally gzip-compressed LIGO-LW XML file containing single-detector
triggers and coincidences.

The root-mean square measurement error depends on the SNR of the signal, so
there is a choice for how to generate perturbed time and phase measurements:

 - `zero-noise`: no measurement error at all
 - `gaussian-noise`: measurement error for a matched filter in Gaussian noise
"""

import copy
import functools

import lal
import numpy as np

from . import (
    ArgumentParser, EnableAction, FileType, get_random_parser,
    register_to_xmldoc, write_fileobj)


def parser():
    # Determine list of known detectors for command line arguments.
    available_ifos = sorted(det.frDetector.prefix
                            for det in lal.CachedDetectors)

    parser = ArgumentParser(parents=[get_random_parser()])
    parser.add_argument(
        'input', metavar='IN.xml[.gz]', type=FileType('rb'),
        default='-', help='Name of input file')
    parser.add_argument(
        '-o', '--output', metavar='OUT.xml[.gz]', type=FileType('wb'),
        default='-', help='Name of output file')
    parser.add_argument(
        '-j', '--jobs', type=int, default=1, const=None, nargs='?',
        help='Number of threads')
    parser.add_argument(
        '--detector', metavar='|'.join(available_ifos), nargs='+',
        help='Detectors to use', choices=available_ifos, required=True)
    parser.add_argument(
        '--waveform',
        help='Waveform to use for injections (overrides values in '
        'sim_inspiral table)')
    parser.add_argument(
        '--snr-threshold', type=float, default=4.,
        help='Single-detector SNR threshold')
    parser.add_argument(
        '--net-snr-threshold', type=float, default=12.,
        help='Network SNR threshold')
    parser.add_argument(
        '--keep-subthreshold', action='store_true',
        help='Keep sub-threshold triggers that do not contribute to the '
        'network SNR')
    parser.add_argument(
        '--min-triggers', type=int, default=2,
        help='Emit coincidences only when at least this many triggers '
        'are found')
    parser.add_argument(
        '--min-distance', type=float, default=0.0,
        help='Skip events with distance less than this value')
    parser.add_argument(
        '--max-distance', type=float, default=float('inf'),
        help='Skip events with distance greater than this value')
    parser.add_argument(
        '--measurement-error', default='zero-noise',
        choices=('zero-noise', 'gaussian-noise'),
        help='How to compute the measurement error')
    parser.add_argument(
        '--enable-snr-series', action=EnableAction, default=True,
        help='Enable output of SNR time series')
    parser.add_argument(
        '--reference-psd', metavar='PSD.xml[.gz]', type=FileType('rb'),
        required=True, help='Name of PSD file')
    parser.add_argument(
        '--f-low', type=float,
        help='Override low frequency cutoff found in sim_inspiral table')
    parser.add_argument(
        '--f-high', type=float,
        help='Set high frequency cutoff to simulate early warning')
    parser.add_argument(
        '--duty-cycle', type=float, default=1.0,
        help='Single-detector duty cycle')
    parser.add_argument(
        '-P', '--preserve-ids', action='store_true',
        help='Preserve original simulation_ids')
    return parser


def simulate_snr(ra, dec, psi, inc, distance, epoch, gmst, H, S,
                 response, location, measurement_error='zero-noise',
                 duration=0.1):
    from scipy.interpolate import interp1d

    from ..bayestar import filter
    from ..bayestar.interpolation import interpolate_max

    # Calculate whitened template autocorrelation sequence.
    HS = filter.signal_psd_series(H, S)
    n = len(HS.data.data)
    acor, sample_rate = filter.autocorrelation(HS, duration)

    # Calculate time, amplitude, and phase.
    u = np.cos(inc)
    u2 = np.square(u)
    signal_model = filter.SignalModel(HS)
    horizon = signal_model.get_horizon_distance()
    Fplus, Fcross = lal.ComputeDetAMResponse(response, ra, dec, psi, gmst)
    toa = lal.TimeDelayFromEarthCenter(location, ra, dec, epoch)
    z = (0.5 * (1 + u2) * Fplus + 1j * u * Fcross) * horizon / distance

    # Calculate complex autocorrelation sequence.
    snr_series = z * np.concatenate((acor[:0:-1].conj(), acor))

    # If requested, add noise.
    if measurement_error == 'gaussian-noise':
        sigmasq = 4 * np.sum(HS.deltaF * np.abs(HS.data.data))
        amp = 4 * n * HS.deltaF**0.5 * np.sqrt(HS.data.data / sigmasq)
        N = lal.CreateCOMPLEX16FrequencySeries(
            '', HS.epoch, HS.f0, HS.deltaF, HS.sampleUnits, n)
        N.data.data = amp * (
            np.random.randn(n) + 1j * np.random.randn(n))
        noise_term, sample_rate_2 = filter.autocorrelation(
            N, len(snr_series) / sample_rate, normalize=False)
        assert sample_rate == sample_rate_2
        snr_series += noise_term

    # Shift SNR series to the nearest sample.
    int_samples, frac_samples = divmod(
        (1e-9 * epoch.gpsNanoSeconds + toa) * sample_rate, 1)
    if frac_samples > 0.5:
        int_samples += 1
        frac_samples -= 1
    epoch = lal.LIGOTimeGPS(epoch.gpsSeconds, 0)
    n = len(acor) - 1
    mprime = np.arange(-n, n + 1)
    m = mprime + frac_samples
    re, im = (
        interp1d(m, x, kind='cubic', bounds_error=False, fill_value=0)(mprime)
        for x in (snr_series.real, snr_series.imag))
    snr_series = re + 1j * im

    # Find the trigger values.
    i_nearest = np.argmax(np.abs(snr_series[n-n//2:n+n//2+1])) + n-n//2
    i_interp, z_interp = interpolate_max(i_nearest, snr_series,
                                         n // 2, method='lanczos')
    toa = epoch + (int_samples + i_interp - n) / sample_rate
    snr = np.abs(z_interp)
    phase = np.angle(z_interp)

    # Shift and truncate the SNR time series.
    epoch += (int_samples + i_nearest - n - n // 2) / sample_rate
    snr_series = snr_series[(i_nearest - n // 2):(i_nearest + n // 2 + 1)]
    tseries = lal.CreateCOMPLEX8TimeSeries(
        'snr', epoch, 0, 1 / sample_rate,
        lal.DimensionlessUnit, len(snr_series))
    tseries.data.data = snr_series
    return horizon, snr, phase, toa, tseries


def simulate(seed, sim_inspiral, psds, responses, locations, measurement_error,
             f_low=None, f_high=None, waveform=None):
    from ..bayestar import filter

    np.random.seed(seed)

    # Unpack some values from the row in the table.
    DL = sim_inspiral.distance
    ra = sim_inspiral.longitude
    dec = sim_inspiral.latitude
    inc = sim_inspiral.inclination
    # phi = sim_inspiral.coa_phase  # arbitrary?
    psi = sim_inspiral.polarization
    epoch = sim_inspiral.time_geocent
    gmst = lal.GreenwichMeanSiderealTime(epoch)

    f_low = f_low or sim_inspiral.f_lower
    waveform = waveform or sim_inspiral.waveform

    # Signal models for each detector.
    H = filter.sngl_inspiral_psd(
        waveform,
        mass1=sim_inspiral.mass1,
        mass2=sim_inspiral.mass2,
        spin1x=sim_inspiral.spin1x,
        spin1y=sim_inspiral.spin1y,
        spin1z=sim_inspiral.spin1z,
        spin2x=sim_inspiral.spin2x,
        spin2y=sim_inspiral.spin2y,
        spin2z=sim_inspiral.spin2z,
        f_min=f_low,
        f_final=f_high)

    return [
        simulate_snr(
            ra, dec, psi, inc, DL, epoch, gmst, H, S, response, location,
            measurement_error=measurement_error)
        for S, response, location in zip(psds, responses, locations)]


def main(args=None):
    p = parser()
    with p.parse_args(args) as opts:
        # LIGO-LW XML imports.
        from ligo.lw import ligolw
        from ligo.lw.param import Param
        from ligo.lw.utils.search_summary import append_search_summary
        from ligo.lw import utils as ligolw_utils
        from ligo.lw.lsctables import (
            New, CoincDefTable, CoincID, CoincInspiralTable, CoincMapTable,
            CoincTable, ProcessParamsTable, ProcessTable, SimInspiralTable,
            SnglInspiralTable, TimeSlideTable)

        # glue, LAL and pylal imports.
        from ligo import segments
        import lal
        import lal.series
        import lalsimulation
        from lalinspiral.inspinjfind import InspiralSCExactCoincDef
        from lalinspiral.thinca import InspiralCoincDef
        from tqdm import tqdm

        # BAYESTAR imports.
        from ..io.events.ligolw import ContentHandler
        from ..bayestar import filter
        from ..util.progress import progress_map

        # Read PSDs.
        xmldoc = ligolw_utils.load_fileobj(
            opts.reference_psd, contenthandler=lal.series.PSDContentHandler)
        psds = lal.series.read_psd_xmldoc(xmldoc, root_name=None)
        psds = {
            key: filter.InterpolatedPSD(filter.abscissa(psd), psd.data.data)
            for key, psd in psds.items() if psd is not None}
        psds = [psds[ifo] for ifo in opts.detector]

        # Extract simulation table from injection file.
        inj_xmldoc = ligolw_utils.load_fileobj(
            opts.input, contenthandler=ContentHandler)
        orig_sim_inspiral_table = SimInspiralTable.get_table(inj_xmldoc)

        # Prune injections that are outside distance limits.
        orig_sim_inspiral_table[:] = [
            row for row in orig_sim_inspiral_table
            if opts.min_distance <= row.distance <= opts.max_distance]

        # Open output file.
        xmldoc = ligolw.Document()
        xmlroot = xmldoc.appendChild(ligolw.LIGO_LW())

        # Create tables. Process and ProcessParams tables are copied from the
        # injection file.
        coinc_def_table = xmlroot.appendChild(New(CoincDefTable))
        coinc_inspiral_table = xmlroot.appendChild(New(CoincInspiralTable))
        coinc_map_table = xmlroot.appendChild(New(CoincMapTable))
        coinc_table = xmlroot.appendChild(New(CoincTable))
        xmlroot.appendChild(ProcessParamsTable.get_table(inj_xmldoc))
        xmlroot.appendChild(ProcessTable.get_table(inj_xmldoc))
        sim_inspiral_table = xmlroot.appendChild(New(SimInspiralTable))
        sngl_inspiral_table = xmlroot.appendChild(New(SnglInspiralTable))
        time_slide_table = xmlroot.appendChild(New(TimeSlideTable))

        # Write process metadata to output file.
        process = register_to_xmldoc(
            xmldoc, p, opts, instruments=opts.detector,
            comment="Simulated coincidences")

        # Add search summary to output file.
        all_time = segments.segment([lal.LIGOTimeGPS(0), lal.LIGOTimeGPS(2e9)])
        append_search_summary(xmldoc, process, inseg=all_time, outseg=all_time)

        # Create a time slide entry.  Needed for coinc_event rows.
        time_slide_id = time_slide_table.get_time_slide_id(
            {ifo: 0 for ifo in opts.detector}, create_new=process)

        # Populate CoincDef table.
        inspiral_coinc_def = copy.copy(InspiralCoincDef)
        inspiral_coinc_def.coinc_def_id = coinc_def_table.get_next_id()
        coinc_def_table.append(inspiral_coinc_def)
        found_coinc_def = copy.copy(InspiralSCExactCoincDef)
        found_coinc_def.coinc_def_id = coinc_def_table.get_next_id()
        coinc_def_table.append(found_coinc_def)

        # Precompute values that are common to all simulations.
        detectors = [lalsimulation.DetectorPrefixToLALDetector(ifo)
                     for ifo in opts.detector]
        responses = [det.response for det in detectors]
        locations = [det.location for det in detectors]

        func = functools.partial(simulate, psds=psds,
                                 responses=responses, locations=locations,
                                 measurement_error=opts.measurement_error,
                                 f_low=opts.f_low, f_high=opts.f_high,
                                 waveform=opts.waveform)

        # Make sure that each thread gets a different random number state.
        # We start by drawing a random integer s in the main thread, and
        # then the i'th subprocess will seed itself with the integer i + s.
        #
        # The seed must be an unsigned 32-bit integer, so if there are n
        # threads, then s must be drawn from the interval [0, 2**32 - n).
        #
        # Note that *we* are thread 0, so there are a total of
        # n=1+len(sim_inspiral_table) threads.
        seed = np.random.randint(0, 2 ** 32 - len(sim_inspiral_table) - 1)
        np.random.seed(seed)

        with tqdm(desc='accepted') as progress:
            for sim_inspiral, simulation in zip(
                    orig_sim_inspiral_table,
                    progress_map(
                        func,
                        np.arange(len(orig_sim_inspiral_table)) + seed + 1,
                        orig_sim_inspiral_table, jobs=opts.jobs)):

                sngl_inspirals = []
                used_snr_series = []
                net_snr = 0.0
                count_triggers = 0

                # Loop over individual detectors and create SnglInspiral
                # entries.
                for ifo, (horizon, abs_snr, arg_snr, toa, series) \
                        in zip(opts.detector, simulation):

                    if np.random.uniform() > opts.duty_cycle:
                        continue
                    elif abs_snr >= opts.snr_threshold:
                        # If SNR < threshold, then the injection is not found.
                        # Skip it.
                        count_triggers += 1
                        net_snr += np.square(abs_snr)
                    elif not opts.keep_subthreshold:
                        continue

                    # Create SnglInspiral entry.
                    used_snr_series.append(series)
                    sngl_inspirals.append(
                        sngl_inspiral_table.RowType(**dict(
                            dict.fromkeys(
                                sngl_inspiral_table.validcolumns, None),
                            process_id=process.process_id,
                            ifo=ifo,
                            mass1=sim_inspiral.mass1,
                            mass2=sim_inspiral.mass2,
                            spin1x=sim_inspiral.spin1x,
                            spin1y=sim_inspiral.spin1y,
                            spin1z=sim_inspiral.spin1z,
                            spin2x=sim_inspiral.spin2x,
                            spin2y=sim_inspiral.spin2y,
                            spin2z=sim_inspiral.spin2z,
                            end=toa,
                            snr=abs_snr,
                            coa_phase=arg_snr,
                            f_final=opts.f_high,
                            eff_distance=horizon / abs_snr)))

                net_snr = np.sqrt(net_snr)

                # If too few triggers were found, then skip this event.
                if count_triggers < opts.min_triggers:
                    continue

                # If network SNR < threshold, then the injection is not found.
                # Skip it.
                if net_snr < opts.net_snr_threshold:
                    continue

                # Add Coinc table entry.
                coinc = coinc_table.appendRow(
                    coinc_event_id=coinc_table.get_next_id(),
                    process_id=process.process_id,
                    coinc_def_id=inspiral_coinc_def.coinc_def_id,
                    time_slide_id=time_slide_id,
                    insts=opts.detector,
                    nevents=len(opts.detector),
                    likelihood=None)

                # Add CoincInspiral table entry.
                coinc_inspiral_table.appendRow(
                    coinc_event_id=coinc.coinc_event_id,
                    instruments=[
                        sngl_inspiral.ifo for sngl_inspiral in sngl_inspirals],
                    end=lal.LIGOTimeGPS(1e-9 * np.mean([
                        sngl_inspiral.end.ns()
                        for sngl_inspiral in sngl_inspirals
                        if sngl_inspiral.end is not None])),
                    mass=sim_inspiral.mass1 + sim_inspiral.mass2,
                    mchirp=sim_inspiral.mchirp,
                    combined_far=0.0,  # Not provided
                    false_alarm_rate=0.0,  # Not provided
                    minimum_duration=None,  # Not provided
                    snr=net_snr)

                # Record all sngl_inspiral records and associate them with
                # coincs.
                for sngl_inspiral, series in zip(
                        sngl_inspirals, used_snr_series):
                    # Give this sngl_inspiral record an id and add it to the
                    # table.
                    sngl_inspiral.event_id = sngl_inspiral_table.get_next_id()
                    sngl_inspiral_table.append(sngl_inspiral)

                    if opts.enable_snr_series:
                        elem = lal.series.build_COMPLEX8TimeSeries(series)
                        elem.appendChild(
                            Param.from_pyvalue(
                                'event_id', sngl_inspiral.event_id))
                        xmlroot.appendChild(elem)

                    # Add CoincMap entry.
                    coinc_map_table.appendRow(
                        coinc_event_id=coinc.coinc_event_id,
                        table_name=sngl_inspiral_table.tableName,
                        event_id=sngl_inspiral.event_id)

                # Record injection
                if not opts.preserve_ids:
                    sim_inspiral.simulation_id \
                        = sim_inspiral_table.get_next_id()
                sim_inspiral_table.append(sim_inspiral)

                progress.update()

        # Record coincidence associating injections with events.
        for i, sim_inspiral in enumerate(sim_inspiral_table):
            coinc = coinc_table.appendRow(
                coinc_event_id=coinc_table.get_next_id(),
                process_id=process.process_id,
                coinc_def_id=found_coinc_def.coinc_def_id,
                time_slide_id=time_slide_id,
                instruments=None,
                nevents=None,
                likelihood=None)
            coinc_map_table.appendRow(
                coinc_event_id=coinc.coinc_event_id,
                table_name=sim_inspiral_table.tableName,
                event_id=sim_inspiral.simulation_id)
            coinc_map_table.appendRow(
                coinc_event_id=coinc.coinc_event_id,
                table_name=coinc_table.tableName,
                event_id=CoincID(i))

        # Record process end time.
        process.set_end_time_now()

        # Write output file.
        write_fileobj(xmldoc, opts.output)



# FIXME: Remove this file if https://github.com/willvousden/ptemcee/pull/6
# is merged
import numpy as np
import ptemcee.sampler

__all__ = ('Sampler',)


class VectorLikePriorEvaluator(ptemcee.sampler.LikePriorEvaluator):

    def __call__(self, x):
        s = x.shape
        x = x.reshape((-1, x.shape[-1]))

        lp = self.logp(x, *self.logpargs, **self.logpkwargs)
        if np.any(np.isnan(lp)):
            raise ValueError('Prior function returned NaN.')

        ll = np.empty_like(lp)
        bad = (lp == -np.inf)
        ll[bad] = 0
        ll[~bad] = self.logl(x[~bad], *self.loglargs, **self.loglkwargs)
        if np.any(np.isnan(ll)):
            raise ValueError('Log likelihood function returned NaN.')

        return ll.reshape(s[:-1]), lp.reshape(s[:-1])


class Sampler(ptemcee.sampler.Sampler):
    """Patched version of :class:`ptemcee.Sampler` that supports the
    `vectorize` option of :class:`emcee.EnsembleSampler`.
    """

    def __init__(self, nwalkers, dim, logl, logp,  # noqa: N803
                 ntemps=None, Tmax=None, betas=None,  # noqa: N803
                 threads=1, pool=None, a=2.0,
                 loglargs=[], logpargs=[],
                 loglkwargs={}, logpkwargs={},
                 adaptation_lag=10000, adaptation_time=100,
                 random=None, vectorize=False):
        super().__init__(nwalkers, dim, logl, logp,
                         ntemps=ntemps, Tmax=Tmax, betas=betas,
                         threads=threads, pool=pool, a=a, loglargs=loglargs,
                         logpargs=logpargs, loglkwargs=loglkwargs,
                         logpkwargs=logpkwargs, adaptation_lag=adaptation_lag,
                         adaptation_time=adaptation_time, random=random)
        self._vectorize = vectorize
        if vectorize:
            self._likeprior = VectorLikePriorEvaluator(logl, logp,
                                                       loglargs, logpargs,
                                                       loglkwargs, logpkwargs)

    def _evaluate(self, ps):
        if self._vectorize:
            return self._likeprior(ps)
        else:
            return super().evaluate(ps)



#
# Copyright (C) 2018-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import functools
import numpy as np
from numpy._core.umath import _add_newdoc_ufunc

__all__ = ('add_newdoc_ufunc', 'require_contiguous_aligned')


def add_newdoc_ufunc(func, doc):  # pragma: no cover
    """Set the docstring for a Numpy ufunc.

    The function :func:`numpy._core.umath._add_newdoc_ufunc` can only change a
    ufunc's docstring if it is `NULL`. This workaround avoids an exception when
    the user tries to `reload()` this module.

    Notes
    -----
    :func:`numpy._core.umath._add_newdoc_ufunc` is not part of Numpy's public
    API, but according to upstream developers it is unlikely to go away any
    time soon.

    See https://github.com/numpy/numpy/issues/26233.
    """
    try:
        _add_newdoc_ufunc(func, doc)
    except ValueError as e:
        msg = 'Cannot change docstring of ufunc with non-NULL docstring'
        if e.args[0] == msg:
            pass


def require_contiguous_aligned(func):
    """Wrap a Numpy ufunc to guarantee that all of its inputs are
    C-contiguous arrays.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        n = func.nin
        args = [arg if i >= n or np.isscalar(arg)
                else np.require(arg, requirements={'CONTIGUOUS', 'ALIGNED'})
                for i, arg in enumerate(args)]
        return func(*args, **kwargs)
    return wrapper



#
# Copyright (C) 2011-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Plot an all-sky map on a Mollweide projection.
By default, plot in celestial coordinates (RA, Dec).

To plot in geographic coordinates (longitude, latitude) with major
coastlines overlaid, provide the ``--geo`` flag.

Public-domain cartographic data is courtesy of `Natural Earth
<http://www.naturalearthdata.com>`_ and processed with `MapShaper
<http://www.mapshaper.org>`_.
"""

from . import ArgumentParser, FileType, SQLiteType
from .matplotlib import get_figure_parser


def parser():
    parser = ArgumentParser(parents=[get_figure_parser()])
    parser.add_argument(
        '--annotate', default=False, action='store_true',
        help='annotate plot with information about the event')
    parser.add_argument(
        '--contour', metavar='PERCENT', type=float, nargs='+',
        help='plot contour enclosing this percentage of'
        ' probability mass [may be specified multiple times, default: none]')
    parser.add_argument(
        '--colorbar', default=False, action='store_true',
        help='Show colorbar')
    parser.add_argument(
        '--radec', nargs=2, metavar='deg', type=float, action='append',
        default=[], help='right ascension (deg) and declination (deg) to mark')
    parser.add_argument(
        '--inj-database', metavar='FILE.sqlite', type=SQLiteType('r'),
        help='read injection positions from database')
    parser.add_argument(
        '--geo', action='store_true',
        help='Use a terrestrial reference frame with coordinates (lon, lat) '
             'instead of the celestial frame with coordinates (RA, dec) '
             'and draw continents on the map')
    parser.add_argument(
        '--projection', type=str, default='mollweide',
        choices=['mollweide', 'aitoff', 'globe', 'zoom'],
        help='Projection style')
    parser.add_argument(
        '--projection-center', metavar='CENTER',
        help='Specify the center for globe and zoom projections, e.g. 14h 10d')
    parser.add_argument(
        '--zoom-radius', metavar='RADIUS',
        help='Specify the radius for zoom projections, e.g. 4deg')
    parser.add_argument(
        'input', metavar='INPUT.fits[.gz]', type=FileType('rb'),
        default='-', nargs='?', help='Input FITS file')
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        # Late imports
        import numpy as np
        import matplotlib.pyplot as plt
        from matplotlib import rcParams
        from ..io import fits
        from .. import moc
        from .. import plot
        from .. import postprocess
        from astropy.coordinates import SkyCoord
        from astropy.table import Table
        from astropy.time import Time
        from astropy import units as u

        sr_to_deg2 = u.sr.to(u.deg**2)

        skymap = fits.read_sky_map(opts.input.name, moc=True)

        axes_args = {}
        if opts.geo:
            axes_args['projection'] = 'geo'
            obstime = Time(skymap.meta['gps_time'], format='gps').utc.isot
            axes_args['obstime'] = obstime
        else:
            axes_args['projection'] = 'astro'
        axes_args['projection'] += ' ' + opts.projection
        if opts.projection_center is not None:
            axes_args['center'] = SkyCoord(opts.projection_center)
        if opts.zoom_radius is not None:
            axes_args['radius'] = opts.zoom_radius
        ax = plt.axes(**axes_args)
        ax.grid()

        # Add contours.
        if opts.contour:
            dA = moc.uniq2pixarea(skymap['UNIQ'])
            dP = skymap['PROBDENSITY'] * dA
            cls = 100 * postprocess.find_greedy_credible_levels(
                dP, skymap['PROBDENSITY'])

            table = Table({'UNIQ': skymap['UNIQ'], 'CLS': cls})
            cs = ax.contour_hpx(
                (table, 'ICRS'), colors='k', linewidths=0.5,
                levels=opts.contour, order='nearest-neighbor')
            fmt = r'%g\%%' if rcParams['text.usetex'] else '%g%%'
            plt.clabel(cs, fmt=fmt, fontsize=6, inline=True)

        # Add continents.
        if opts.geo:
            plt.plot(*plot.coastlines(), color='0.5', linewidth=0.5,
                     transform=ax.get_transform('world'))

        radecs = opts.radec
        if opts.inj_database:
            query = '''SELECT DISTINCT longitude, latitude
                       FROM sim_inspiral AS si
                       INNER JOIN coinc_event_map AS cm1
                       ON (si.simulation_id = cm1.event_id)
                       INNER JOIN coinc_event_map AS cm2
                       ON (cm1.coinc_event_id = cm2.coinc_event_id)
                       WHERE cm2.event_id = ?
                       AND cm1.table_name = 'sim_inspiral'
                       AND cm2.table_name = 'coinc_event'
                    '''
            (ra, dec), = opts.inj_database.execute(
                query, (skymap.meta['objid'],)).fetchall()
            radecs.append(np.rad2deg([ra, dec]).tolist())

        # Add markers (e.g., for injections or external triggers).
        for ra, dec in radecs:
            ax.plot_coord(
                SkyCoord(ra, dec, unit='deg'), '*',
                markerfacecolor='white', markeredgecolor='black',
                markersize=10)

        # Add a white outline to all text to make it stand out from the
        # background.
        plot.outline_text(ax)

        if opts.annotate:
            text = []
            try:
                objid = skymap.meta['objid']
            except KeyError:
                pass
            else:
                text.append('event ID: {}'.format(objid))
            if opts.contour:
                i = np.flipud(np.argsort(skymap['PROBDENSITY']))
                areas = np.interp(
                    opts.contour, cls[i], np.cumsum(dA[i]),
                    left=0, right=4*np.pi)
                pp = np.round(opts.contour).astype(int)
                ii = np.round(areas * sr_to_deg2).astype(int)
                for i, p in zip(ii, pp):
                    # FIXME: use Unicode symbol instead of TeX '$^2$'
                    # because of broken fonts on Scientific Linux 7.
                    text.append('{:d}% area: {:,d} deg²'.format(p, i))
            ax.text(1, 1, '\n'.join(text), transform=ax.transAxes, ha='right')

        # Plot sky map.
        skymap['PROBDENSITY'] *= 1 / sr_to_deg2
        img = ax.imshow_hpx((skymap, 'ICRS'), vmin=0, order='nearest-neighbor')

        # Add colorbar.
        if opts.colorbar:
            cb = plot.colorbar(img)
            cb.set_label(r'prob. per deg$^2$')

        # Show or save output.
        opts.output()



import astropy_healpix as ah
from astropy.coordinates import CartesianRepresentation, SkyCoord
from astropy.table import Table
from astropy import units as u
import numpy as np
import healpy as hp
import pytest
from scipy import stats

from ...distance import cartesian_kde_to_moments, moments_to_parameters
from ...moc import nest2uniq
from ..crossmatch import crossmatch


@pytest.fixture
def cartesian_gaussian():
    """Generate a 3D Cartesian Gaussian with random mean and covariance."""
    rotation = stats.special_ortho_group(3).rvs()
    eigenvalues = stats.gamma(2).rvs(3)
    cov = rotation @ np.diag(eigenvalues) @ rotation.T
    mean = stats.norm.rvs(size=3)
    return stats.multivariate_normal(mean, cov)


def cartesian_gaussian_to_skymap(level, mean, cov):
    """Convert a 3D Cartesian Gaussian to a 3D sky map."""
    # Set up HEALPix grid.
    nside = 2**level
    npix = ah.nside_to_npix(nside)
    ipix = np.arange(npix)
    coords = np.column_stack(hp.pix2vec(nside, ipix, nest=True))

    # Create sky map using same method as KDE to convert from Cartesian
    # to spherical representation. This is just a special case where there
    # is only a single cluster and a single KDE sample.
    means = mean[np.newaxis, ..., np.newaxis]
    inv_covs = np.linalg.inv(cov)[np.newaxis, ...]
    weights = np.ones(1)
    probdensity, distmean, diststd = np.transpose([
        cartesian_kde_to_moments(n, means, inv_covs, weights) for n in coords])

    # Create 3D, multi-order sky map.
    uniq = nest2uniq(level, ipix)
    distmu, distsigma, distnorm = moments_to_parameters(distmean, diststd)
    return Table({'UNIQ': uniq,
                  'PROBDENSITY': probdensity,
                  'DISTMU': distmu,
                  'DISTSIGMA': distsigma,
                  'DISTNORM': distnorm})


@pytest.fixture(params=range(3))
def contours(request):
    """Generate random values for the ``contours`` argument."""
    return np.random.uniform(size=request.param)


@pytest.mark.parametrize('n_coordinates', [None, *range(3)])
def test_crossmatch_cartesian_gaussian_distribution(
        cartesian_gaussian, contours, n_coordinates):
    """Test on a Cartesian Gaussian distribution.

    This distribution has closed-form expressions for the following outputs:
    * contour_vols
    * probdensity_vol
    * searched_prob_vol
    * searched_vol
    """
    skymap = cartesian_gaussian_to_skymap(
        6, cartesian_gaussian.mean, cartesian_gaussian.cov)

    if n_coordinates is None:
        coordinates = None
    else:
        coordinates_xyz = cartesian_gaussian.rvs(size=n_coordinates)
        coordinates = SkyCoord(*coordinates_xyz.T * u.Mpc,
                               representation_type=CartesianRepresentation)

    result = crossmatch(skymap, contours=contours, coordinates=coordinates)

    standard_vol = 4/3*np.pi * np.sqrt(np.linalg.det(cartesian_gaussian.cov))
    expected = standard_vol * stats.chi(3).ppf(contours)**3
    np.testing.assert_allclose(result.contour_vols, expected, rtol=2e-3)

    if coordinates is None:
        assert np.isnan(result.probdensity_vol)
        assert np.isnan(result.searched_prob_vol)
        assert np.isnan(result.searched_vol)
    elif np.size(coordinates) == 0:
        assert np.size(result.probdensity_vol) == 0
        assert np.size(result.searched_prob_vol) == 0
        assert np.size(result.searched_vol) == 0
    else:
        expected = cartesian_gaussian.pdf(coordinates_xyz)
        np.testing.assert_allclose(result.probdensity_vol, expected, rtol=4e-2)

        d = coordinates_xyz - cartesian_gaussian.mean
        r = np.sqrt(np.sum(((d @ np.linalg.inv(cartesian_gaussian.cov)) * d),
                           axis=-1))
        expected = stats.chi(3).cdf(r)
        np.testing.assert_allclose(result.searched_prob_vol, expected,
                                   atol=1e-2)

        expected = standard_vol * r**3
        np.testing.assert_allclose(result.searched_vol, expected, rtol=6e-2)

        # FIXME: to calculate the expected value for result.searched_prob_dist,
        # one would need the cdf of a generalized chi square distribution.
        # There is not a convenient Python implementation, but there is in R:
        # https://cran.r-project.org/web/packages/CompQuadForm/index.html.
        # See literature references in that project's PDF documentation.



#
# Copyright (C) 2013-2024  Giuseppe Greco, Leo Singer, and CDS team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Create a contour for a credible level of an all-sky probability map. The input
is a HEALPix FITS probability map. The output is a `Multi-Order Coverage (MOC)
<http://ivoa.net/documents/MOC/>`_ FITS file.
"""

from . import ArgumentParser, FileType


def parser():
    parser = ArgumentParser()

    parser.add_argument(
        '-o', '--output', metavar='FILE.fits', required=True,
        help='output file')
    parser.add_argument(
        '-c', '--contour', metavar='PERCENT', type=float, required=True,
        help='MOC region enclosing this percentage of probability \
              [range is 0-100]')
    parser.add_argument(
        'input', metavar='INPUT.fits[.gz]', type=FileType('rb'),
        default='-', nargs='?', help='Input multi-order or flatten \
                                      HEALPix FITS file')

    return parser


def main(args=None):
    p = parser()
    with parser().parse_args(args) as opts:
        import astropy_healpix as ah
        import astropy.units as u

        try:
            from mocpy import MOC
        except ImportError:
            p.error('This command-line tool requires mocpy >= 0.8.2. '
                    'Please install it by running "pip install mocpy".')

        from ..io import read_sky_map

        # Read multi-order sky map
        skymap = read_sky_map(opts.input.name, moc=True)

        uniq = skymap['UNIQ']
        probdensity = skymap['PROBDENSITY']

        level, ipix = ah.uniq_to_level_ipix(uniq)
        area = ah.nside_to_pixel_area(
            ah.level_to_nside(level)).to_value(u.steradian)

        prob = probdensity * area

        # Create MOC
        contour_decimal = opts.contour / 100
        moc = MOC.from_valued_healpix_cells(
            uniq, prob, max_depth=level.max(),
            cumul_from=0.0, cumul_to=contour_decimal)

        # Write MOC
        moc.write(opts.output, format='fits', overwrite=True)



# This file is adapted from the Astropy package template, which is licensed
# under a 3-clause BSD style license - see licenses/TEMPLATE_LICENSE.rst

__all__ = ['__version__']

# this indicates whether or not we are in the package's setup.py
try:
    _ASTROPY_SETUP_
except NameError:
    import builtins
    builtins._ASTROPY_SETUP_ = False

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''


if not _ASTROPY_SETUP_:  # noqa
    import os

    # Create the test function for self test
    from astropy.tests.runner import TestRunner
    test = TestRunner.make_test_runner_in(os.path.dirname(__file__))
    test.__test__ = False
    __all__ += ['test']



from astropy.coordinates import EarthLocation, ITRS, SkyCoord
from astropy.time import Time
from astropy import units as u
from pytest import approx

from .. import DetectorFrame


def test_detector_frame():
    site_1 = EarthLocation.of_site('H1')
    site_2 = EarthLocation.of_site('L1')
    t = Time(1e9, format='gps')

    detector_frame = DetectorFrame(site_1=site_1, site_2=site_2, obstime=t)
    itrs_frame = ITRS(obstime=t)

    itrs_coord = SkyCoord(*(u.Quantity(site_1.geocentric) -
                            u.Quantity(site_2.geocentric)).value,
                          frame=itrs_frame)
    assert itrs_coord.transform_to(detector_frame).lat.deg == 90

    detector_coord = SkyCoord(lon=0 * u.deg, lat=90 * u.deg,
                              frame=detector_frame)
    converted = detector_coord.transform_to(itrs_frame)
    assert converted.spherical.lon.deg == approx(itrs_coord.spherical.lon.deg)
    assert converted.spherical.lat.deg == approx(itrs_coord.spherical.lat.deg)



# Copyright (C) 2017-2020  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Modify events by artificially disabling specified detectors."""
from .base import Event, EventSource

__all__ = ('DetectorDisabledEventSource', 'DetectorDisabledError')


class DetectorDisabledError(ValueError):
    pass


class DetectorDisabledEventSource(EventSource):

    def __init__(self, base_source, disabled_detectors, raises=True):
        self.base_source = base_source
        self.disabled_detectors = set(disabled_detectors)
        self.raises = raises

    def __iter__(self):
        return iter(self.base_source)

    def __getitem__(self, key):
        return DetectorDisabledEvent(self, self.base_source[key])

    def __len__(self):
        return len(self.base_source)


class DetectorDisabledEvent(Event):

    def __init__(self, source, base_event):
        self.source = source
        self.base_event = base_event

    @property
    def singles(self):
        disabled_detectors = self.source.disabled_detectors
        if self.source.raises:
            detectors = {s.detector for s in self.base_event.singles}
            if not detectors & disabled_detectors:
                raise DetectorDisabledError(
                    'Disabling detectors {{{}}} would have no effect on this '
                    'event with detectors {{{}}}'.format(
                        ' '.join(sorted(disabled_detectors)),
                        ' '.join(sorted(detectors))))
            if not detectors - disabled_detectors:
                raise DetectorDisabledError(
                    'Disabling detectors {{{}}} would exclude all data for '
                    'this event with detectors {{{}}}'.format(
                        ' '.join(sorted(disabled_detectors)),
                        ' '.join(sorted(detectors))))
        return tuple(s for s in self.base_event.singles
                     if s.detector not in disabled_detectors)

    @property
    def template_args(self):
        return self.base_event.template_args


open = DetectorDisabledEventSource



#
# Copyright (C) 2013-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Produce GW sky maps for all coincidences in search pipeline output database
in LIGO-LW XML, LIGO-LW SQLite, or PyCBC HDF5 format.

The distance prior is controlled by the ``--prior-distance-power`` argument.
If you set ``--prior-distance-power`` to k, then the distance prior is
proportional to r^k. The default is 2, uniform in volume.

If the ``--min-distance`` argument is omitted, it defaults to zero. If the
``--max-distance argument`` is omitted, it defaults to the SNR=4 horizon
distance of the most sensitive detector.

A FITS file is created for each sky map, having a filename of the form
``X.fits`` where X is the integer LIGO-LW row ID of the coinc. The ``OBJECT``
card in the FITS header is also set to the integer row ID.
"""

from . import (
    ArgumentParser, FileType, get_waveform_parser, get_posterior_parser,
    get_mcmc_parser, get_random_parser)


ROW_ID_COMMENT = [
    '',
    'The integer value in the OBJECT card in this FITS header is a row ID',
    'that refers to a coinc_event table row in the input LIGO-LW document.',
    '']


def parser():
    parser = ArgumentParser(
        parents=[get_waveform_parser(), get_posterior_parser(),
                 get_mcmc_parser(), get_random_parser()])
    parser.add_argument(
        '-d', '--disable-detector', metavar='X1', type=str, nargs='+',
        help='disable certain detectors')
    parser.add_argument(
        '--keep-going', '-k', default=False, action='store_true',
        help='Keep processing events if a sky map fails to converge')
    parser.add_argument(
        'input', metavar='INPUT.{hdf,xml,xml.gz,sqlite}', default='-',
        nargs='+', type=FileType('rb'),
        help='Input LIGO-LW XML file, SQLite file, or PyCBC HDF5 files. '
             'For PyCBC, you must supply the coincidence file '
             '(e.g. "H1L1-HDFINJFIND.hdf" or "H1L1-STATMAP.hdf"), '
             'the template bank file (e.g. H1L1-BANK2HDF.hdf), '
             'the single-detector merged PSD files '
             '(e.g. "H1-MERGE_PSDS.hdf" and "L1-MERGE_PSDS.hdf"), '
             'and the single-detector merged trigger files '
             '(e.g. "H1-HDF_TRIGGER_MERGE.hdf" and '
             '"L1-HDF_TRIGGER_MERGE.hdf"), '
             'in any order.')
    parser.add_argument(
        '--pycbc-sample', default='foreground',
        help='(PyCBC only) sample population')
    parser.add_argument(
        '--coinc-event-id', type=int, nargs='*',
        help='run on only these specified events')
    parser.add_argument(
        '--output', '-o', default='.',
        help='output directory')
    parser.add_argument(
        '--condor-submit', action='store_true',
        help='submit to Condor instead of running locally')
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        import logging
        log = logging.getLogger('BAYESTAR')

        # BAYESTAR imports.
        from .. import omp
        from ..io import fits, events
        from ..bayestar import localize

        # Other imports.
        import os
        import subprocess
        import sys

        log.info('Using %d OpenMP thread(s)', omp.num_threads)

        # Read coinc file.
        log.info(
            '%s:reading input files', ','.join(
                file.name for file in opts.input))
        event_source = events.open(*opts.input, sample=opts.pycbc_sample)

        if opts.disable_detector:
            event_source = events.detector_disabled.open(
                event_source, opts.disable_detector)

        os.makedirs(opts.output, exist_ok=True)

        if opts.condor_submit:
            if opts.seed is not None:
                raise NotImplementedError(
                    '--seed does not yet work with --condor-submit')
            if opts.coinc_event_id:
                raise ValueError(
                    'must not set --coinc-event-id with --condor-submit')
            with subprocess.Popen(['condor_submit'],
                                  text=True, stdin=subprocess.PIPE) as proc:
                f = proc.stdin
                print('''
                    accounting_group = ligo.dev.o4.cbc.pe.bayestar
                    on_exit_remove = (ExitBySignal == False) && (ExitCode == 0)
                    on_exit_hold = (ExitBySignal == True) || (ExitCode != 0)
                    on_exit_hold_reason = (ExitBySignal == True \
                        ? strcat("The job exited with signal ", ExitSignal) \
                        : strcat("The job exited with code ", ExitCode))
                    request_memory = 2000 MB
                    request_disk = 100 MB
                    universe = vanilla
                    getenv = true
                    executable = /usr/bin/env
                    JobBatchName = BAYESTAR
                    environment = "OMP_NUM_THREADS=1"
                    ''', file=f)
                print(
                    'error =', os.path.join(opts.output, '$(cid).err'), file=f)
                print(
                    'arguments = "',
                    *(arg for arg in sys.argv if arg != '--condor-submit'),
                    '--coinc-event-id $(cid)"', file=f)
                print('queue cid in', *event_source, file=f)
            sys.exit(proc.returncode)

        if opts.coinc_event_id:
            event_source = {
                key: event_source[key] for key in opts.coinc_event_id}

        count_sky_maps_failed = 0

        # Loop over all sngl_inspiral <-> sngl_inspiral coincs.
        for coinc_event_id, event in event_source.items():
            # Loop over sky localization methods
            log.info('%d:computing sky map', coinc_event_id)
            if opts.chain_dump:
                chain_dump = f'{coinc_event_id}.hdf5'
            else:
                chain_dump = None
            try:
                sky_map = localize(
                    event, opts.waveform, opts.f_low, opts.min_distance,
                    opts.max_distance, opts.prior_distance_power,
                    opts.cosmology, mcmc=opts.mcmc, chain_dump=chain_dump,
                    enable_snr_series=opts.enable_snr_series,
                    f_high_truncate=opts.f_high_truncate,
                    rescale_loglikelihood=opts.rescale_loglikelihood)
                sky_map.meta['objid'] = coinc_event_id
                sky_map.meta['comment'] = ROW_ID_COMMENT

            except (ArithmeticError, ValueError):
                log.exception('%d:sky localization failed', coinc_event_id)
                count_sky_maps_failed += 1
                if not opts.keep_going:
                    raise
            else:
                log.info('%d:saving sky map', coinc_event_id)
                filename = f'{coinc_event_id}.fits'
                fits.write_sky_map(
                    os.path.join(opts.output, filename), sky_map, nest=True)

        if count_sky_maps_failed > 0:
            raise RuntimeError("{0} sky map{1} did not converge".format(
                count_sky_maps_failed,
                's' if count_sky_maps_failed > 1 else ''))



from importlib import metadata
from importlib.resources import as_file, files
import multiprocessing
import os
import shutil
import stat
import subprocess
import sys
import tempfile

dist = 'ligo.skymap'
group = 'console_scripts'

__all__ = ('entry_points', 'run_entry_point', 'run_ligolw', 'run_lalsuite')

entry_points = {entry_point.name: entry_point
                for entry_point in metadata.distribution(dist).entry_points
                if entry_point.group == group}


def run_entry_point(name, *args):
    main = entry_points[name].load()
    try:
        main(args)
    except SystemExit as e:
        if e.code != 0:
            raise subprocess.CalledProcessError(e.code, [name, *args])


def exec_ligolw(name, *args):
    main, = (entry_point.load() for entry_point
             in metadata.distribution('python-ligo-lw').entry_points
             if entry_point.name == name)
    sys.argv = [name, *args]
    main()


def run_ligolw(name, *args):
    """Run an external tool that is provided by python-ligo-lw.

    This is trivial if python-ligo-lw has actually been installed and LALSuite
    is in the PATH. If python-ligo-lw is not installed and is only present as
    an egg, then things get more complicated.
    """
    path = shutil.which(name)
    if path:
        # The tool has been installed, so we can just call it.
        subprocess.check_call([path, *args])
    else:
        # The tool has not been installed, so we have to try to run it using
        # importlib.metadata.
        process = multiprocessing.Process(
            target=exec_ligolw, args=[name, *args])
        process.start()
        process.join()
        if process.exitcode != 0:
            raise subprocess.CalledProcessError(
                process.exitcode, [name, *args])


def run_lalsuite(name, *args):
    """Run an external tool that is provided by LALSuite.

    This is trivial if lalsuite has actually been installed it is in the PATH.
    If LALSuite is not installed and is only present as an egg, then things get
    more complicated because of how the LALSuite wheel packs binaries as
    package data.
    """
    path = shutil.which(name)
    if path:
        # The tool has been installed, so we can just call it.
        subprocess.check_call([path, *args])
    else:
        # The tool has not been installed, so we have to find the underlying
        # binary inside the lalapps module.
        with as_file(files('lalapps.bin').joinpath(name)) as path:
            # Copy to a temporary file so that we can make it executable.
            # For some reason, when eggs are extracted, permissions are not
            # preserved.
            with tempfile.NamedTemporaryFile(dir=os.path.dirname(path)) as tmp:
                with open(path, 'rb') as orig:
                    shutil.copyfileobj(orig, tmp)
                tmp.flush()
                fd = tmp.fileno()
                stat_result = os.fstat(fd)
                mode = stat_result.st_mode | stat.S_IXUSR
                os.fchmod(fd, mode)
                subprocess.check_call([tmp.name, *args])



#
# Copyright (C) 2023-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import astropy_healpix as ah
import numpy as np
from reproject.utils import parse_output_projection
from reproject.healpix.utils import parse_coord_system

from ..healpix_tree import HEALPIX_MACHINE_ORDER, HEALPIX_MACHINE_NSIDE
from ..moc import uniq2nest


def reproject_from_healpix_moc(
    input_data, output_projection, shape_out=None
):
    """
    Reproject multiorder HEALPIX data to a standard projection.
    Adapted from :meth:`reproject.reproject_from_healpix`.

    Parameters
    ----------
    input_data : tuple
        A tuple consisting of the following:
        -   A multi-order HEALPix dataset stored as an Astropy table whose
            first column is called UNIQ and contains the NUNIQ pixel index.
            Every point on the unit sphere must be contained in exactly one
            pixel in the dataset.
        -   An instance of `~astropy.coordinates.BaseCoordinateFrame` or a
            string alias for a coordinate frame.
    output_projection : `~astropy.wcs.WCS` or `~astropy.io.fits.Header`
        The output projection, which can be either a `~astropy.wcs.WCS`
        or a `~astropy.io.fits.Header` instance.
    shape_out : tuple, optional
        If ``output_projection`` is a `~astropy.wcs.WCS` instance, the
        shape of the output data should be specified separately.

    Returns
    -------
    array_new : `~numpy.ndarray`
        The reprojected array.
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values.
    """
    array_in, coord_system_in = input_data
    coord_system_in = parse_coord_system(coord_system_in)
    wcs_out, shape_out = parse_output_projection(
        output_projection, shape_out=shape_out)

    # Look up lon, lat of pixels in reference system and convert celestial
    # coordinates
    yinds, xinds = np.indices(shape_out)
    world_in = wcs_out.pixel_to_world(xinds, yinds).transform_to(
        coord_system_in)
    world_in_cart = world_in.represent_as("cartesian").xyz.value

    hpx_in = ah.xyz_to_healpix(
        *world_in_cart, nside=HEALPIX_MACHINE_NSIDE, order='nested')
    order, hpx_data = uniq2nest(array_in['UNIQ'])
    hpx_data <<= 2 * (HEALPIX_MACHINE_ORDER - order)
    sorter = np.argsort(hpx_data)
    i = np.searchsorted(hpx_data, hpx_in, 'right', sorter=sorter) - 1
    data = array_in.columns[1][sorter][i]

    footprint = (hpx_in != -1).astype(float)

    return data, footprint



# Copyright (C) 2017-2020  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Read events from either HDF or LIGO-LW files."""
import builtins
import sqlite3

from ligo.lw.ligolw import Element
import h5py

from . import hdf, ligolw, sqlite

__all__ = ('MagicEventSource', 'open')


def _read_file_header(f, nbytes=16):
    """Read the first 16 bytes of a file

    This is presumed to include the characters that declare the
    file type.

    Parameters
    ----------
    f : file, str
        A file object or the path to a file

    Returns
    -------
    header : bytes
        A string (hopefully) describing the file type

    """
    try:
        pos = f.tell()
    except AttributeError:
        with builtins.open(f, "rb") as fobj:
            return fobj.read(nbytes)
    try:
        return f.read(nbytes)
    finally:
        f.seek(pos)


def MagicEventSource(f, *args, **kwargs):  # noqa: N802
    """Read events from LIGO-LW XML, LIGO-LW SQlite, or HDF5 files. The format
    is determined automatically using the :manpage:`file(1)` command, and then
    the file is opened using :obj:`.ligolw.open`, :obj:`.sqlite.open`, or
    :obj:`.hdf.open`, as appropriate.

    Returns
    -------
    `~ligo.skymap.io.events.EventSource`

    """
    if isinstance(f, h5py.File):
        opener = hdf.open
    elif isinstance(f, sqlite3.Connection):
        opener = sqlite.open
    elif isinstance(f, Element):
        opener = ligolw.open
    else:
        fileheader = _read_file_header(f)
        if fileheader.startswith(b'\x89HDF\r\n\x1a\n'):
            opener = hdf.open
        elif fileheader.startswith(b'SQLite format 3'):
            opener = sqlite.open
        elif fileheader.startswith((
                b'<?xml',  # XML
                b'\x1f\x8b\x08',  # GZIP
        )):
            opener = ligolw.open
        else:
            raise IOError('Unknown file format')
    return opener(f, *args, **kwargs)


open = MagicEventSource



from time import sleep

import numpy as np
import pytest

from ..progress import progress_map


def func(x):
    sleep(np.random.uniform(0, 0.1))
    return np.square(x)


@pytest.mark.parametrize('jobs', [1, 8])
def test_map(jobs):
    x = np.arange(20)
    result = list(progress_map(func, x, jobs=jobs))
    np.testing.assert_array_equal(result, np.square(x))
